/******************************************************************
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Copyright:
 * Georg Schild (dangertools@gmail.com)
 * Franz Endstrasser (franz.endstrasser@gmail.com)
 ******************************************************************/

package core;

/**
 * ControlThread to be implemented by a specific backend. The implemented ControlThread has to be a
 * Managedthread too cause we are calling methods not directly but via reflection. Errorchecking should
 * be done in spinDown() (from ManagedThread) which is called after executing the rest. You can block
 * execution by doing wait() in enterCritical() and notifyAll() in spinDown() or some other way of waiting
 * for a free connection e.g.
 */

public interface ControlThread {
	
	
	/**
	 * start playback
	 * 
	 * @return		empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue play();
	
	/**
	 * stops playback
	 * 
	 * @return		empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue stop();
	
	/**
	 * pauses playback
	 * 
	 * @return		empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue pause();
	
	/**
	 * plays the next track in the list
	 * 
	 * @return		empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue next();
	
	/**
	 * plays the previous track in the list
	 * 
	 * @return		empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue previous();
	
	/**
	 * seeks the current song per ms milliseconds
	 * 
	 * @param ms		milliseconds to seek
	 * @return			empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue seek(long ms);
	
	/**
	 * Sets the next title in the playlist given by an absolute index and starts playing it
	 * 
	 * @param abs		the next index to play
	 * @return			an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue nextInPlaylistAbs(int abs);
	
	/**
	 * Sets the next title in the playlist given by a relative index and starts playing it
	 * 
	 * @param rel		the next index to play
	 * @return			an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue nextInPlaylistRel(int rel);

	/**
	 * clears the current playlist
	 * 
	 * @return		an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue playlistClear();
	
	/**
	 * shuffles the current playlist
	 * 
	 * @return		an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue playlistShuffle();
	
	/**
	 * Adds a title to the playlist
	 * 
	 * @param path		path to the title given as URL
	 * @return			an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue addTitleToPlaylist(String path);
	
	/**
	 * Adds a title to the playlist
	 * 
	 * @param id		id of the title in the medialib to add
	 * @return			an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue addTitleToPlaylist(int id);

	/**
	 * Adds a title to the playlist 
	 * 
	 * @param id		id of the title in the medialib to add
	 * @param index	index to add the title to
	 * @return			an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue addTitleToPlaylist(int id, int index);
	
	/**
	 * adds a given title to the playlist at a given index
	 * 
	 * @param path		path to the title given as URL 
	 * @param index	index to add the title to
	 * @return			is an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue addTitleToPlaylistIndex(String path, int index);
	
	/**
	 * Removes the given title from the playlist
	 * 
	 * @param index	index of the title
	 * @return			an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue removeTitleFromPlaylist(long index);
	
	/**
	 * Removes the given titles from the playlist
	 * 
	 * @param indices	indices of the titles
	 * @return			an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue removeTitlesFromPlaylist(long indices[]);
	
	/**
	 * Set a given setting on the backend
	 * 
	 * @param setting		the setting to change
	 * @param value		new value of the setting
	 * @return				an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue setSetting(String setting, String value);
	
	/**
	 * Increase or decrease volume of the backend with the given value
	 * 
	 * @param val		relative value of volume +/-[0-100]
	 * @return			an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue creaseVolume(int val);

	/**
	 * Set volume of the backend
	 * 
	 * @param val		val to change the volume to [0-100]
	 * @return			an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue setVolume(int val);

	/**
	 * Set a mixer value
	 * 
	 * @param mixer	number of mixer
	 * @param val		set mixer to
	 * @return			an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue setMixer(int mixer, int val);
	
	/**
	 * Sorts the playlist by a given argument (e.g. artist, title, ...)
	 * 
	 * @param by		sort by
	 * @return			ReturnValue of type VOID_TYPE
	 */
	abstract ReturnValue playlistSort(String by);
}
