/******************************************************************
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Copyright:
 * Georg Schild (dangertools@gmail.com)
 * Franz Endstrasser (franz.endstrasser@gmail.com)
 ******************************************************************/

package core;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Observable;
import java.util.Observer;

import org.apache.log4j.Logger;

/**
 * Every backend has has to extend this KeepStateThread. You can put callbackhandling,
 * computation of signals, mainloop, ... in here. This class has to hold data from backend
 * e.g. volume, current playlist, etc.
 */

public abstract class KeepStateThread extends Observable implements Runnable{
	protected Map configs;
	protected Logger logger;
	protected Management mm;
	private int playbackStatus = 0, backgroundWorkerStatus = 0;
	private Playlist currentPlaylist;
	private int volume = 0;
	
	public static final int IDLE = 0;
	public static final int PLAYING = 1;
	public static final int PAUSED = 2;
	public static final int ERROR = -1;
	
	/**
	 * Constructor creates a new KeepStateThread but doesn't start it.
	 * 
	 * @param ob		the observer gets notified with objects of UpdateUIType
	 * @param mm		Management
	 */
	protected KeepStateThread(Observer ob, Management mm){
		configs = Collections.synchronizedMap(new HashMap());
		logger = Logger.getLogger(getClass());
		this.mm = mm;
		currentPlaylist = new Playlist();
		addObserver(ob);
	}
	
	/**
	 * method is called when spinning down the backend. you could destroy your mainloop here e.g.
	 *
	 */
	protected abstract void spinDown();
	/**
	 * called after backend has come up again
	 * 		
	 */
	protected abstract void restart();
	/**
	 * Return an array of supported mediafile-types
	 * @return	supported types
	 */
	protected abstract String[] getFileTypes();
	/**
	 * Return all supported filetypes including playlists etc.
	 * @return		all supported types
	 */
	protected abstract String[] getAllSupportedTypes();
	/**
	 * Return supported playlisttypes
	 * @return		supported playlisttypes
	 */
	protected abstract String[] getPlaylistTypes();
	/**
	 * get value of equalizer x
	 * @param x
	 * @return		value
	 */
	protected abstract int getEQ(int x);
	/**
	 * do some things when playlist is getting busy cause we add a bunch of files e.g.
	 * @param flag
	 */
	protected abstract void playlistBusy(boolean flag);
	/**
	 * do some things when medialib is getting busy cause we add a bunch of files e.g.
	 * @param flag
	 */
	protected abstract void mlibBusy(boolean flag);
	
	/**
	 * set playtime to new value and update observers
	 * @param val
	 */
	protected final void updatePlaytime(double val){
		setChanged();
		notifyObservers(new UpdateUIType(UpdateUIType.PLAYTIME_CHANGED, new Double(val)));
	}
	
	/**
	 * Sets current index in playlist [1..n] and fires an event at the observers
	 * 
	 * @param index
	 */
	protected final void setCurrentPlaylistIndex(int index){
		currentPlaylist.setPosition(index);
		setChanged();
		notifyObservers(new UpdateUIType(UpdateUIType.PLAYLIST_CURRENT_POSITION_CHANGED, new Integer(index)));
	}
	
	/**
	 * Sets playbackstatus to one of the defined types and fires an event at the observers
	 * 
	 * @param state
	 */
	protected final void setPlaybackStatus(int state){
		playbackStatus = state;
		setChanged();
		notifyObservers(new UpdateUIType(UpdateUIType.PLAYBACK_STATUS_CHANGED, new Integer(state)));
	}
	
	/**
	 * Set status of backgroundworker job to either BG_IDLE or BG_WORKING
	 * @param val
	 */
	protected final void setBackgroundWorkerStatus(int val){
		backgroundWorkerStatus = val;
		setChanged();
		notifyObservers(new UpdateUIType(UpdateUIType.BG_WORKING, null));
	}
	
	/**
	 * return status of backgroundworker
	 * @return		BG_IDLE if idle, else BG_WORKING
	 */
	protected final int getBackgroundWorkerStatus(){
		return backgroundWorkerStatus;
	}
	
	/**
	 * set volume to new value, used by backend. fires up all observers
	 * @param val
	 */
	protected final void setVolume(int val){
		volume = val;
		setChanged();
		notifyObservers(new UpdateUIType(UpdateUIType.VOLUME_OR_EQUALIZER_CHANGED, new Integer(-1)));
	}
	
	/**
	 * return current volume as value from 1 to 100
	 * @return		current volume
	 */
	protected final int getVolume(){
		return volume;
	}
	
	/**
	 * returns the current playlist
	 * 
	 * @return		Playlist
	 */
	protected final Playlist getCurrentPlaylist(){
		return currentPlaylist;
	}
	
	/**
	 * Set current playlist and fire event. used by backend
	 * @param current
	 */
	protected final void setCurrentPlaylist(Playlist current){
		currentPlaylist = current;
		setChanged();
		notifyObservers(new UpdateUIType(UpdateUIType.PLAYLIST_CHANGED, new Object[]{}));
	}
	
	/**
	 * Update a specific configvalue and fire an event at the observer
	 * 
	 * @param config
	 * @param value
	 */
	protected final void updateConfig(String config, String value){
		if (configs.containsKey(config))
			configs.put(config,value);
		else logger.debug("don't know configval " + config);
		setChanged();
		notifyObservers(new UpdateUIType(UpdateUIType.CONFIGVAL_CHANGED, config));
	}
	
	/**
	 * return current index in playlist
	 * @return		current index
	 */
	protected final int getCurrentPlaylistIndex(){
		return currentPlaylist.getPosition();
	}
	
	/**
	 * return playbackstatus of backend
	 * 
	 * @return		playbackstatus
	 */
	protected final int getPlaybackStatus(){
		return playbackStatus;
	}
	
	/**
	 * return Map with all configs
	 * @return		configs from backend
	 */
	protected final Map getConfigs(){
		return configs;
	}
}
