/******************************************************************
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Copyright:
 * Georg Schild (dangertools@gmail.com)
 * Franz Endstrasser (franz.endstrasser@gmail.com)
 ******************************************************************/

package core;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Observable;
import java.util.Observer;
import java.util.Timer;
import java.util.TimerTask;

import org.apache.log4j.Logger;

import statics.StaticValues;

/**
 * Class is extended by <Backendname>ControlThread and <Backendname>MedialibThread. Its methods are 
 * used to run methods from the interfaces via reflection
 */
public abstract class ManagedThread extends Observable implements Runnable {
	protected ReturnValue ret;
	protected Method invokeMethod;
	protected Object[] methodArgs;
	protected Logger logger;
	protected Management mm;
	private Thread t;
	
	/**
	 * Create new ManagedThread with given values
	 * 
	 * @param ret
	 * @param obs
	 * @param mm
	 */
	protected ManagedThread(ReturnValue ret, Observer obs, Management mm){
		this.ret = ret;
		addObserver(obs);
		this.mm = mm;
		logger = Logger.getLogger(getClass());
	}
	
	/**
	 * used internally to get the right Method
	 * @param name
	 * @param args
	 * @return			Method for name
	 * @throws NoSuchMethodException	if the method isn't found
	 */
	private Method getMethodByName(String name, Object[] args) throws NoSuchMethodException{
		Method[] ms = getClass().getDeclaredMethods();
		for ( int i = 0; i < ms.length; i++){
			if (ms[i].getName().equals(name)){
				boolean check = true;
				Class[] types = ms[i].getParameterTypes();
				if (types.length != args.length)
					check = false;
				for ( int j = 0; j < types.length && check; j++){
					if (args[j] != null){
						String second = args[j].getClass().getName();
						if (second.equals("java.lang.Integer"))
							second = "int";
						else if (second.equals("java.lang.Long"))
							second = "long";
						else if (second.equals("java.lang.Boolean"))
							second = "boolean";
						else if (second.equals("java.lang.Double"))
							second = "double";
						if (!types[j].equals(args[j].getClass()) && !types[j].getName().equals(second))
							check = false;
					}
					else check = false;
				}
				if (check)
					return ms[i];
			}
		}
		throw new NoSuchMethodException();
	}
	
	/**
	 * Return tid of this Thread
	 * @return		tid
	 */
	public final long getThreadID(){
		return ret.getThreadID();
	}
	
	/**
	 * run the method set by runMethod(), spinDown() and notify observers afterwards
	 */
	public final void run() {
		try {
			while (!mm.isConnected())
				Thread.sleep(100);
			if (!enterCritical()){
				logger.fatal("Thread didn't accept execution!!\nMethod was: " + invokeMethod + "\n" +
						"Arguments where: " + methodArgs + "\nThread was a " + getClass());
				return;
			}
			Timer clock = new Timer();
			clock.schedule(new TimerTask(){
				public void run() {
					try{
						if (t.isAlive()){
							t.interrupt();
						}
					} catch (Exception e){
						logger.debug(e.getMessage(), e);
					}
				}
			}, 15000);
			Object o = invokeMethod.invoke(this, methodArgs);
			clock.cancel();
			String error = spinDown();
			if ( error != null ){
				logger.warn("Error occured in backend: "+error);
				ret.setValue(error);
			}
			setChanged();
			notifyObservers(o);
		} catch (InvocationTargetException i){
			logger.info(StaticValues.getLangString("ReflectionMethodException") + " \"" + 
					invokeMethod + "\": " + StaticValues.getLangString("targetNotUseable"));
			logger.debug(i.getMessage(), i);
		} catch (IllegalAccessException i){
			logger.info(StaticValues.getLangString("ReflectionMethodException") + " \"" + 
					invokeMethod + "\": " + StaticValues.getLangString("noAccess"));
			logger.debug(i.getMessage(), i);
		} catch (InterruptedException e) {
			logger.debug(StaticValues.getLangString("sleepInterrupt"));
			logger.debug(e.getMessage(), e);
		}
	}
	
	/**
	 * run a method and notify observers after that.
	 * 
	 * @param iMethod		name of the method to be run (just "play", "pause", ..., not the whole name)
	 * @param args			arguments to the method
	 */
	public final void runMethod(String iMethod, Object[] args){
		try {
			invokeMethod = getMethodByName(iMethod, args);
		}catch (NoSuchMethodException n){
			logger.info(StaticValues.getLangString("ReflectionMethodException") + " \"" + 
					iMethod + "\": " + StaticValues.getLangString("noSuchMethod"));
			logger.debug(n.getMessage(), n);
			return;
		}
		methodArgs = args;
		t = new Thread(this);
		t.start();
	}
	
	public final String toString(){
		return "[" + getThreadID() + "]";
	}
	
	/**
	 * Called when spinning down the thread.
	 * 
	 * @return		return a String with some error text or "" if nothing happened
	 */
	protected abstract String spinDown();
	
	/**
	 * Do whatever things need to be done to enter the critical section but be careful when
	 * releasing such a lock later on, e.g. in spinDown()
	 * 
	 * @return		if true we'll go and execute the thing, otherwise we stop here
	 */
	protected abstract boolean enterCritical();
}
