/******************************************************************
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Copyright:
 * Georg Schild (dangertools@gmail.com)
 * Franz Endstrasser (franz.endstrasser@gmail.com)
 ******************************************************************/

package core;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Observable;
import java.util.Observer;
import java.util.Timer;
import java.util.TimerTask;

import org.Xwidgets.Styles;
import org.Xwidgets.WidgetFactory;
import org.Xwidgets.XMessageBox;
import org.Xwidgets.SWT.SWTFactory;
import org.apache.log4j.Logger;

import core.xmms2.Xmms2ThreadFactory;

import statics.StaticValues;

/**
 * Management is the interface between Controller (usually UI) and backend. If a method returns a long-value
 * that is a threadID cause the underlying code starts one of Medialib- or Controlthread. Otherwise we get
 * the values directly from the KeepStateThread or from some other persistent source
 */
public final class Management extends Observable implements Observer {
	private Map returnedThreads, titleCache;
	private ThreadFactory threadFac;
	private long threadID = 1000;
	private Logger logger;
	private KeepStateThread kst;
	private static Management instance;
	private boolean upAndRunning = false;
	
	private Management(){
		logger = Logger.getLogger(getClass());
		returnedThreads = Collections.synchronizedMap(new HashMap());
		titleCache = Collections.synchronizedMap(new HashMap());
		initBackend();
		if (threadFac != null){
			kst = threadFac.getKeepStateThread(this);
		}
		new Timer().scheduleAtFixedRate(new TimerTask(){
			public void run() {
				synchronized(returnedThreads){
					for (Iterator it = returnedThreads.values().iterator(); it.hasNext();){
						ReturnValue rv = (ReturnValue)it.next();
						if ((rv.getTimeStamp()+11000) < System.currentTimeMillis())
							it.remove();
					}
				}
				synchronized(titleCache){
					for (Iterator it = titleCache.values().iterator(); it.hasNext();){
						Title t = (Title)it.next();
						if (t != null && (Long.parseLong(t.getAttribute("startTime"))+700000) < System.currentTimeMillis())
							it.remove();
					}
				}
				System.gc();
			}
			
		}, 10000, 10000);
		upAndRunning = true;
	}
	
	/**
	 * 
	 * @return		return true if everything is up and running, false otherwise
	 */
	public boolean upAndRunning(){
		return upAndRunning; 
	}
	
	/**
	 * init backend and handle exceptions got from it at init-time
	 *
	 */
	private void initBackend(){
		if (StaticValues.getSetting("backend").equals("xmms2"))
			try {
				threadFac = new Xmms2ThreadFactory(this);
			} catch (Exception e) {
				WidgetFactory mine;
				try {
					mine = SWTFactory.getInstance();
					XMessageBox error = mine.getMessageBox(null, Styles.ERROR|Styles.OK);
					error.setText(StaticValues.getLangString("backendInitProblem") + ":\n" + e.getLocalizedMessage());
					error.open();
					System.exit(1);
				} catch (Exception e1) {
					logger.debug("outer: " + e.getMessage(), e);
					logger.debug("inner: " + e1.getMessage(), e1);
				}
			}
		else
			logger.warn("backend \"" + StaticValues.getSetting("backend") + "\" is not known");
	}
	
	/**
	 * get an instance of the Management
	 * 
	 * @param ob		observer (usually a Controller)
	 * @return			return the Management instance
	 */
	public static Management getInstance(Observer ob){
		if (instance == null)
			instance = new Management();
		instance.addObserver(ob);
		return instance;
	}
	/* The following Methods are used internal and are not directly connected to a Thread
	 * from backend
	 */
	/**
	 * get an instance of the Management
	 * 
	 * @return			return the Management instance
	 */
	public static Management getInstance(){
		if (instance == null)
			instance = new Management();
		return instance;
	}
	/**
	 * 
	 * @return		true if we are connected to the backend, false otherwise
	 */
	public boolean isConnected(){
		return threadFac.isConnected();
	}
	/**
	 * Update all data
	 *
	 */
	public void updateAll(){
		/*if (kst != null)
			kst.updateAll();*/
	}
	/**
	 * called on spinDown to do things
	 *
	 */
	public void spinDown(){
		threadFac.spinDown();
	}
	/**
	 * check if the given id is returned already. if you wait too long the Sweeper has already got
	 * the thread and killed it
	 * 
	 * @param ID
	 * @return			ReturnValue or null
	 */
	public ReturnValue isReturned(long ID){
		return (ReturnValue)returnedThreads.get(new Long(ID));
	}
	/**
	 * 
	 * @return		default SQL String backend uses for title-queries
	 */
	public String getDefaultSQLString(){
		return ThreadFactory.defaultSQLString; 
	}
	/**
	 * get the next available threadID
	 * 
	 * @return		threadID
	 */
	private long getNextThreadID(){
		return ++threadID;
	}
	/**
	 * restart backend (after crash e.g.)
	 *
	 */
	public void restartBackend(){
		if (kst != null)
			kst.restart();
	}
	/**
	 * waits for max 6s on a given id
	 * 
	 * @param id
	 * @return			ReturnValue got from isReturned or null
	 */
	public ReturnValue waitForID(long id){
		ReturnValue ret = null;
		int cnt = 0;
		while ((ret = isReturned(id)) == null && cnt++ < 60)
			try {
				Thread.sleep(40);
			} catch (InterruptedException e) {
				logger.debug(StaticValues.getLangString("sleepInterrupt"));
			}
		return ret;
	}
	/**
	 * update the Management or the Controller which is Observer from Management
	 * 
	 * - if the method receives an object of type ReturnValue we will take it and work directly with it
	 * (maybe wrap it to a UpdateUIType)
	 * - if the method receives an object of type UpdateUIType we will forward it to our observers
	 */
	public void update(Observable arg0, Object arg1) {
		if (arg1 instanceof ReturnValue){
			ReturnValue r = (ReturnValue)arg1;
			if (r.getType() == ReturnValue.MEDIALIB_TYPE){
				UpdateUIType uit = new UpdateUIType(UpdateUIType.MLIB_LIST, r.getValue());
				setChanged();
				notifyObservers(uit);
			}
			else if (r.getType() == ReturnValue.MEDIALIB_SEARCH_TYPE){
				UpdateUIType uit = new UpdateUIType(UpdateUIType.MLIB_SEARCH, r.getValue());
				setChanged();
				notifyObservers(uit);
			}
			else if (r.getType() == ReturnValue.PLAYLISTINFORMATION_TYPE){
				UpdateUIType uit = new UpdateUIType(UpdateUIType.PLAYLIST_INFO, r.getValue());
				setChanged();
				notifyObservers(uit);
			}
			else if (r.getType() == ReturnValue.TITLE_TYPE){
				try {
					Title t = (Title)r.getValue();
					t.setAttribute("startTime", ""+System.currentTimeMillis());
					titleCache.put(new Integer(t.getAttribute("id")), t);
				} catch (Exception e){
					logger.debug(e.getMessage(), e);
				}
			}
			returnedThreads.put(new Long(r.getThreadID()), r);
		}
		else if (arg1 instanceof UpdateUIType){
			setChanged();
			notifyObservers(arg1);
		}
	}
	/**
	 * run a ControlThread which should execute the given method. That Method is not synchronized. If you
	 * need synchronization in your backend you have to handle it on your own
	 * 
	 * @param methodName		name of the method (only "play", "pause", ..., not the whole name)
	 * 		if Methodname starts with "mlib" we'll create a Medialibthread, Controlthread otherwise
	 * @param args			args needed for the method
	 * @return				threadID or -1 if something stupid happened somewhere
	 */
	private long runThread(String methodName, Object[] args){
		ReturnValue t = new ReturnValue(getNextThreadID(), System.currentTimeMillis());
		ManagedThread mt = null;
		if (methodName.startsWith("mlib")){
			mt = threadFac.getMedialibThread(t, this);
			if (mt != null){
				upAndRunning = true;
				mt.runMethod(methodName, args);
				return t.getThreadID();
			}
		}
		else {
			mt = threadFac.getControlThread(t, this);
			if (mt != null){
				upAndRunning = true;
				mt.runMethod(methodName, args);
				return t.getThreadID();
			}
		}
		returnedThreads.put(new Long(t.getThreadID()), t);
		return -1;
	}
	
	
	
	
	/*
	 * Following methods are done via ControlThread. All methods return a threadID
	 */
	/**
	 * play the current playlist index
	 * 
	 * @return		threadID
	 */
	public long play(){
		return runThread("play", new Object[]{});
	}
	/**
	 * stop playback
	 * 
	 * @return		threadID
	 */
	public long stop(){
		return runThread("stop", new Object[]{});
	}
	/**
	 * pause playback if currently playing or resume if already paused
	 * 
	 * @return		threadID
	 */
	public long pause(){
		return runThread("pause", new Object[]{});
	}
	/**
	 * play the next title in playlist or do nothing if already at the end
	 * 
	 * @return		threadID
	 */
	public long next(){
		return runThread("next", new Object[]{});
	}
	/**
	 * play the previous title in the playlist or restart the current if the first one
	 * 
	 * @return		threadID
	 */
	public long prev(){
		return runThread("previous", new Object[]{});
	}
	/**
	 * play the title at playlistindex index
	 * 
	 * @param index	index to play
	 * @return			threadID
	 */
	public long nextInPlaylistAbs(int index){
		return runThread("nextInPlaylistAbs", new Integer[]{new Integer(index)});
	}
	/**
	 * play the title at playlistindex+index
	 * 
	 * @param index	index to be added to currentindex
	 * @return			threadID
	 */
	public long nextInPlaylistRel(int index){
		return runThread("nextInPlaylistRel", new Integer[]{new Integer(index)});
	}
	/**
	 * Shuffle playlist
	 * @return		threadID
	 */
	public long playlistShuffle() {
		return runThread("playlistShuffle", new Object[]{});
	}
	/**
	 * Sort playlist by a given argument
	 * @param sorter 
	 * @return		threadID
	 */
	public long playlistSort(String sorter) {
		return runThread("playlistSort", new Object[]{sorter});
	}
	/**
	 * increase volume by value
	 * 
	 * @param value
	 * @return			threadID
	 */
	public long increaseVolume(int value){
		return runThread("creaseVolume", new Integer[]{new Integer(value)});
	}
	/**
	 * decrease volume by value
	 * 
	 * @param value
	 * @return			threadID
	 */
	public long decreaseVolume(int value){
		return runThread("creaseVolume", new Integer[]{new Integer(-value)});
	}
	/**
	 * set volume to val
	 * 
	 * @param val
	 * @return		threadID
	 */
	public long setVolume(int val){
		return runThread("setVolume", new Integer[]{new Integer(val)});
	}
	/**
	 * set equalizer x to value val 
	 * @param x
	 * @param val
	 * @return			threadID
	 */
	public long setEQ(int x, int val){
		return runThread("setMixer", new Integer[]{new Integer(x), new Integer(val)});
	}
	/**
	 * seek playback by seconds
	 * 
	 * @param seconds		
	 * @return				threadID
	 */
	public long seek(int seconds){
		return runThread("seek", new Long[]{new Long(seconds*1000)});
	}
	/**
	 * add title to playlist
	 * @param url
	 * @return			threadID
	 */
	public long addTitleToPlaylist(String url){
		return runThread("addTitleToPlaylist", new String[]{url});
	}
	/**
	 * add title with id id to playlist
	 * @param id
	 * @return		threadID
	 */
	public long addIdToPlaylist(int id) {
		return runThread("addTitleToPlaylist", new Integer[]{new Integer(id)});
	}
	/**
	 * adds title with id id to current playlist on index index
	 * @param id
	 * @param index
	 * @return			threadID
	 */
	public long addIdToPlaylist(int id, int index) {
		return runThread("addTitleToPlaylist", new Integer[]{new Integer(id), new Integer(index)});
	}
	/**
	 * removed title at index from playlist
	 * @param index
	 * @return			threadID
	 */
	public long removeTitleFromPlaylist(long index){
		return runThread("removeTitleFromPlaylist", new Long[]{new Long(index)});
	}
	/**
	 * removed title at index from playlist
	 * @param indices
	 * @return			threadID
	 */
	public long removeTitlesFromPlaylist(long indices[]){
		return runThread("removeTitlesFromPlaylist", new Object[]{indices});
	}
	/**
	 * Set configval to new value
	 * @param key
	 * @param value
	 * @return			threadID
	 */
	public long setConfigVal(String key, String value){
		return runThread("setSetting", new String[]{key, value});
	}
	/**
	 * Clear current playlist
	 * @return		threadID
	 */
	public long playlistClear(){
		return runThread("playlistClear", new Object[]{});
	}
	/**
	 * add list of URLs (filenames) to current playlist
	 * calls KeepStateThread.playlistBusy()
	 * 
	 * @param urls
	 */
	public void addURLListToPlaylist(List urls){
		kst.playlistBusy(true);
		for (Iterator it = urls.iterator(); it.hasNext();)
			addTitleToPlaylist(""+it.next());
		kst.playlistBusy(false);
	}
	
	
	
	
	/*
	 * Following methods are done via KeepStateThread and return directly
	 */
	/**
	 * 
	 * @return		playbackstatus of backend (PLAYING, PAUSED, STOPPED)
	 */
	public int getPlaybackStatus(){
		if (kst == null) return -1;
		return kst.getPlaybackStatus();
	}
	/**
	 * 
	 * @return		backgroundworkerstatus (BG_IDLE or BG_WORKING)
	 */
	public int getBackgroundWorkerStatus(){
		if (kst == null) return -1;
		return kst.getBackgroundWorkerStatus();
	}
	/**
	 * 
	 * @param x
	 * @return			value of equalizer x or -51 if there's no equalizer x
	 */
	public int getEQ(int x){
		if (kst == null) return -51;
		return kst.getEQ(x);
	}
	/**
	 * 
	 * @return			supported mediafiletypes
	 */
	public String[] getFileTypes(){
		if (kst == null) return new String[]{""};
		return kst.getFileTypes();
	}
	/**
	 * 
	 * @return		all supported filetypes (incl. playlists etc.)
	 */
	public String[] getAllSupportedTypes(){
		if (kst == null) return new String[]{""};
		return kst.getAllSupportedTypes();
	}
	/**
	 * 
	 * @return			supported playlisttypes
	 */
	public String[] getPlaylistTypes(){
		if (kst == null) return new String[]{""};
		return kst.getPlaylistTypes();
	}
	/**
	 * get the current volume
	 * 
	 * @return		volume as int
	 */
	public int getVolume(){
		if (kst == null) return 0;
		return kst.getVolume();
	}
	/**
	 * return the playbackstatus of the backend (one of the static values in core.ControlThread)
	 *  
	 * @return		playbackstatus
	 */
	public int getStatus(){
		if (kst == null) return -1;
		return kst.getPlaybackStatus();
	}
	/**
	 * return pairs of configname-configval
	 * 
	 * @return		HashMap containing pairs of configname-configval
	 */
	public String[] getConfigs(){
		if (kst == null) return new String[1];
		String ret[] = new String[kst.configs.size()];
		int i = 0;
		for(Iterator it = kst.configs.keySet().iterator(); it.hasNext(); )
			ret[i++] = "" + it.next();
		return ret;
	}
	/**
	 * get configval with key key
	 * @param key
	 * @return		value of configval key
	 */
	public String getConfigVal(String key){
		try {
			return "" + kst.configs.get(key);
		}catch (Exception e){
			logger.debug("Unable to find matching Setting for key " + key);
			return "";
		}
	}
	/**
	 * return current playlistindex (first index is 1, not 0)
	 * 
	 * @return			playlistindex
	 */
	public int getPlaylistIndex(){
		if (kst == null) return 0;
		return kst.getCurrentPlaylistIndex();
	}
	/**
	 * @return		current playing Title
	 */
	public Title getPlayingTitle(){
		if (kst == null) return new Title();
		return kst.getCurrentPlaylist().getCurrentTitle(); 
	}
	/**
	 * returns the current Playlist
	 * 
	 * @return		Playlist
	 */
	public Playlist getCurrentPlaylist(){
		if (kst == null) return new Playlist();
		return kst.getCurrentPlaylist();
	}
	/**
	 * @param index
	 * @return			Title at playlistindex index
	 */
	public Title getTitleForIndex(int index) {
		if (kst == null) return new Title();
		return kst.getCurrentPlaylist().getTitleForIndex(index);
	}
	/**
	 * Get Title for ID id - This method uses a cache to speed up things
	 * 
	 * @param id
	 * @return			Title
	 */
	public Title getTitleForID(int id) {
		Title t = null;
		if (isTitleInCache(id)){
			t = (Title)titleCache.get(new Integer(id));
			t.setAttribute("startTime", ""+System.currentTimeMillis());
		}
		return t;
	}
	public boolean isTitleInCache(int id){
		boolean contains = titleCache.containsKey(new Integer(id));
		if (!contains){
			titleCache.put(new Integer(id), null);
			runThread("mlibGetTitleForID", new Integer[]{new Integer(id)});
		}
		else if (titleCache.get(new Integer(id)) == null){
			return false;
		}
		return contains;
	}
	/**
	 * Method updates a title in cache with new data from medialib
	 * @param id
	 */
	public void updateTitleCache(int id){
		runThread("mlibGetTitleForID", new Integer[]{new Integer(id)});
	}
	
	

	
	/*
	 * Following methods work using the MedialibThread. All return their threadID
	 */
	/**
	 * Select something from medialib
	 * @param arg0		[]{key,value,exact}+
	 * @param search
	 * @return			threadID
	 */
	public long mlibSelect(String[] arg0, boolean search) {
		return runThread("mlibSelect", new Object[]{arg0, new Boolean(search)});
	}
	/**
	 * add entry to medialib
	 * @param url
	 * @return			threadID
	 */
	public long mlibAddEntry(String url) {
		return runThread("mlibAddEntry", new String[]{url});
	}
	/**
	 * update Titleinformation of Title t. Medialibthread knows to do with it
	 * 
	 * @param t
	 * @return		threadID
	 */
	public long mlibUpdateTitle(Title t){
		return runThread("mlibUpdateTitle", new Title[]{t});
	}
	/**
	 * Re-get information for medialib (could take some time) 
	 * @return		threadID
	 */
	public long mlibRefresh(){
		return runThread("mlibRefresh", new Object[]{});
	}
	/**
	 * get the list of playlists available from medialib
	 * @return		threadID
	 */
	public long mlibGetPlaylists() {
		return runThread("mlibGetPlaylists", new Object[]{});
	}
	/**
	 * Save current playlist to name name
	 * @param name
	 * @return			threadID
	 */
	public long mlibSaveCurrentPlaylist(String name) {
		return runThread("mlibSaveCurrentPlaylist", new String[]{name});
	}
	/**
	 * Load playlist with given name from medialib
	 * @param name
	 * @return		threadID
	 */
	public long mlibLoadPlaylist(String name) {
		return runThread("mlibLoadPlaylist", new String[]{name});
	}
	/**
	 * Load playlist from url and store it with given name to medialib
	 * @param url
	 * @param name
	 * @return			threadID
	 */
	public long mlibLoadPlaylist(String url, String name) {
		return runThread("mlibLoadPlaylist", new String[]{url, name});
	}
	/**
	 * Add named playlist to current one
	 * @param name
	 * @return			threadID
	 */
	public long mlibAddPlaylistToCurrent(String name) {
		return runThread("mlibAddPlaylistToCurrent", new String[]{name});
	}
	/**
	 * Add SQL-query to playlist
	 * @param sql
	 * @return			threadID
	 */
	public long mlibAddQueryToPlaylist(String sql) {
		return runThread("mlibAddQueryToPlaylist", new String[]{sql});
	}
	/**
	 * Remove playlist from medialib
	 * @param name
	 * @return		threadID
	 */
	public long mlibRemovePlaylist(String name) {
		return runThread("mlibDeletePlaylist", new String[]{name});
	}
	/**
	 * Rename playlist
	 * @param nameold
	 * @param namenew
	 * @return			threadID
	 */
	public long mlibRenamePlaylist(String nameold, String namenew) {
		return runThread("mlibRenamePlaylist", new String[]{nameold, namenew});
	}
	/**
	 * create new empty playlist
	 * @param name
	 * @return			threaID
	 */
	public long mlibNewPlaylist(String name) {
		return runThread("mlibNewPlaylist", new String[]{name});
	}
	/**
	 * add list of URLs (filenames) to medialib
	 * calls KeepStateThread.mlibBusy()
	 * 
	 * @param urls
	 */
	public void addURLListToMedialib(List urls){
		kst.mlibBusy(true);
		for (Iterator it = urls.iterator(); it.hasNext();)
			mlibAddEntry(""+it.next());
		kst.mlibBusy(false);
	}
}
