/******************************************************************
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Copyright:
 * Georg Schild (dangertools@gmail.com)
 * Franz Endstrasser (franz.endstrasser@gmail.com)
 ******************************************************************/

package core;

/**
 * MedialibThread to be implemented by a specific backend. The implemented MedialibThread has to be a
 * Managedthread too cause we are calling methods not directly but via reflection. Errorchecking should
 * be done in spinDown() (from ManagedThread) which is called after executing the rest. You can block
 * execution by doing wait() in enterCritical() and notifyAll() in spinDown() or some other way of waiting
 * for a free connection e.g.
 */

public interface MedialibThread {
	
	/**
	 * loads the playlist specified through its name
	 * @param name 
	 * 
	 * @return an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue mlibLoadPlaylist(String name);
	
	/**
	 * loads a playlist from a given url
	 * 
	 * @param url
	 * @param name
	 * @return			ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue mlibLoadPlaylist(String url, String name);
	
	/**
	 * adds the titles from the playlist "name" to the current playlist
	 * @param name 
	 * 
	 * @return an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue mlibAddPlaylistToCurrent(String name);

	/**
	 * save currently loaded playlist to "name"
	 * @param name 
	 * 
	 * @return an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue mlibSaveCurrentPlaylist(String name);
	
	/**
	 * removes the playlist with the name "name"
	 * @param name 
	 * 
	 * @return an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue mlibDeletePlaylist(String name);
	
	/**
	 * renames the playlist with the name "oldname" to "newname"
	 * @param nameold 
	 * @param namenew 
	 * 
	 * @return an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue mlibRenamePlaylist(String nameold, String namenew);
	
	/**
	 * renames the playlist with the name "name"
	 * @param name 
	 * 
	 * @return an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue mlibNewPlaylist(String name);
	
	/**
	 * mlibAddEntry adds a new file with the URL "url" to the medialib
	 * @param url 
	 * 
	 * @return		is an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue mlibAddEntry(String url);
	
	/**
	 * add a sql query to the playlist (used to add n random songs e.g.)
	 * @param sql
	 * @return		an empty ReturnValue with type VOID_TYPE
	 */
	abstract ReturnValue mlibAddQueryToPlaylist(String sql);
	
	/**
	 * 
	 * @param url		search query ([{key}{value}{strict})
	 * @param search	if it's a search we'll handle that different in frontend
	 * @return			ReturnValue containing a List of results
	 */
	abstract ReturnValue mlibSelect(String[] url, boolean search);
	
	/**
	 * same as mlibSelect which gets an array but with only one query
	 * 
	 * @param key		
	 * @param value
	 * @param ex
	 * @param search
	 * @return			ReturnValue containing a List of results
	 */
	abstract ReturnValue mlibSelect(String key, String value, String ex, boolean search);
	
	/**
	 * write your own query
	 * 
	 * @param query
	 * @param search
	 * @return			ReturnValue containing a List of results
	 */
	abstract ReturnValue mlibSelect(String query, boolean search);
	
	/**
	 * get all playlists from medialib and return them
	 * @return			ReturnValue containing a List of PLAYLISTINFORMATION_TYPE
	 */
	abstract ReturnValue mlibGetPlaylists();
	
	/**
	 * gets a Title for a given id
	 * @param id
	 * @return			ReturnValue of type TITLE_TYPE
	 */
	abstract ReturnValue mlibGetTitleForID(int id);
	
	/**
	 * update titleinformation (id3 e.g.)
	 * @param t
	 * @return			empty ReturnValue of type VOID_TYPE
	 */
	abstract ReturnValue mlibUpdateTitle(Title t);
	
	/**
	 * Refresh the medialib
	 * 
	 * @return			empty ReturnValue of type VOID_TYPE
	 */
	abstract ReturnValue mlibRefresh();
}
