/******************************************************************
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Copyright:
 * Georg Schild (dangertools@gmail.com)
 * Franz Endstrasser (franz.endstrasser@gmail.com)
 ******************************************************************/

package core.xmms2;

import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.util.Observer;
import java.util.Set;

import org.farng.mp3.MP3File;
import org.farng.mp3.TagException;
import org.farng.mp3.id3.ID3v1;
import org.xmms2.xmms2bindings.*;

import core.ManagedThread;
import core.Management;
import core.MedialibThread;
import core.PlaylistInformation;
import core.ReturnValue;
import core.Title;
import core.UpdateUIType;

/**
 * To get an idea what the methods do have a look at core.ManagedThread and core.MedialibThread
 */
public final class Xmms2MedialibThread extends ManagedThread implements MedialibThread {
	private SWIGTYPE_p_xmmsc_connection_St myConnection;
	private SWIGTYPE_p_xmmsc_result_St result;
	protected static final String defaultQuery = "SELECT distinct id FROM Media ";
	
	protected Xmms2MedialibThread(ReturnValue ret, Observer obs, Management mm, SWIGTYPE_p_xmmsc_connection_St connection) {
		super(ret, obs, mm);
		myConnection = connection;
	}
	
	public ReturnValue mlibAddEntry(String url) {
		logger.debug("mlibAddEntry");
		result = Xmmsclient.xmmsc_medialib_add_entry(myConnection,url);
		Xmmsclient.xmmsc_result_wait(result);
		return ret;
	}
	
	public ReturnValue mlibAddQueryToPlaylist(String sql){
		logger.debug("SQL add to playlist: " + defaultQuery+sql);
		result = Xmmsclient.xmmsc_medialib_add_to_playlist(myConnection, defaultQuery + sql);
		Xmmsclient.xmmsc_result_wait(result);
		return ret;
	}
	
	public ReturnValue mlibSelect(String key, String value, String ex, boolean search) {	
		return mlibSelect(buildQuery(new String[]{key,value,ex}), search);
	}
	
	/**
	 * Check for errors in the passed result
	 * @param res
	 * @return		if error return true, false otherwise
	 */
	private boolean isError(SWIGTYPE_p_xmmsc_result_St res){
		if (Xmmsclient.xmmsc_result_iserror(res) == 1){
			System.out.println(Xmmsclient.xmmsc_result_get_error(res));
			return true;
		}
		return false;
	}
	
	/**
	 * Build title with id from result
	 * @param res
	 * @param id
	 * @return		hopefully a completely filled Title
	 */
	private Title getTitleForID(SWIGTYPE_p_xmmsc_result_St res, int id){
		Xmmsclient.xmmsc_result_wait(res);
		Title t = new Title();
		t.setAttribute("id", ""+id);
		if (isError(res)) return t;
		try {
			Xmmsclient.xmmsc_result_source_preference_set(res, Xmms2ThreadFactory.sourcePrefence);
			String[] got = new String[1];
			got[0] = "";
			Xmmsclient.xmmsc_result_get_dict_entry_str(res, "title", got);
			if (isError(res)) return t;
			if (got[0] == null || got[0].equals("")){
				if (Xmmsclient.xmmsc_result_get_dict_entry_str(res, "url", got) == 1){
					t.setAttribute("url", got[0]);
					File f = new File(URLDecoder.decode(got[0], "UTF-8"));
					t.setAttribute("title", f.getName().substring(0, f.getName().length()-4));
					t.setAttribute("artist", f.getName().substring(0, f.getName().length()-4));
				}
			}
			else {
				t.setAttribute("title", got[0]);
				if (Xmmsclient.xmmsc_result_get_dict_entry_str(res, "artist", got) == 1)
					t.setAttribute("artist", got[0]);
			}
			if (isError(res)) return t;
			if (Xmmsclient.xmmsc_result_get_dict_entry_str(res, "album", got) == 1)
				t.setAttribute("album", got[0]);
			if (isError(res)) return t;
			if (Xmmsclient.xmmsc_result_get_dict_entry_str(res, "url", got) == 1)
				t.setAttribute("url", got[0]);
			if (isError(res)) return t;
			if (Xmmsclient.xmmsc_result_get_dict_entry_str(res, "genre", got) == 1)
				t.setAttribute("genre", got[0]);
			if (isError(res)) return t;
			if (Xmmsclient.xmmsc_result_get_dict_entry_str(res, "comment", got) == 1)
				t.setAttribute("comment", got[0]);
			if (isError(res)) return t;
			if (Xmmsclient.xmmsc_result_get_dict_entry_str(res, "mime", got) == 1)
				t.setAttribute("mime", got[0]);
			if (isError(res)) return t;
			int length[] = new int[1];
			length[0] = 0;
			if (isError(res)) return t;
			Xmmsclient.xmmsc_result_get_dict_entry_int32(res, "duration", length);
			t.setLength(length[0]/1000);
			if (isError(res)) return t;
			Xmmsclient.xmmsc_result_get_dict_entry_int32(res, "samplerate", length);
			t.setAttribute("samplerate", ""+length[0]);
			if (isError(res)) return t;
			Xmmsclient.xmmsc_result_get_dict_entry_int32(res, "bitrate", length);
			t.setAttribute("bitrate", ""+length[0]);
		}catch (Throwable e){
			logger.debug(e.getMessage(), e);
		}
		return t;
	}
	
	/**
	 * Get Playlistinformation for a given id
	 * 
	 * @param id
	 * @return		new Playlistinformation
	 */
	private PlaylistInformation getPlaylistInformationForID(long id){
		String query = "SELECT entry, count(entry) AS cnt FROM Playlistentries WHERE playlist_id = " + id;
		logger.debug(query);
		int items[] = new int[1];
		String name[] = new String[1];
		name[0] = "";
		
		result = Xmmsclient.xmmsc_medialib_select(myConnection, query);
		Xmmsclient.xmmsc_result_wait(result);
		if (Xmmsclient.xmmsc_result_is_list(result) == 1){
			Xmmsclient.xmmsc_result_list_first(result);
			Xmmsclient.xmmsc_result_get_dict_entry_int32(result,"cnt", items);
		}
		
		return new PlaylistInformation(id, name[0], items[0]);
	}
	
	/**
	 * build the query for mlibSelect according to keys and values
	 */
	private String buildQuery(String[] keyvalue) {
		String myQuery = defaultQuery;
		String equ = " LIKE ";
		if (keyvalue.length == 0)
			return myQuery;
		if ((keyvalue.length >= 3) && (keyvalue[keyvalue.length-1].toLowerCase().equals("own"))) 
			return myQuery + keyvalue[0];
		if ( (keyvalue.length >= 3) && (keyvalue[keyvalue.length-1].toLowerCase().equals("exact"))){
			equ = " = ";
			keyvalue[1] = keyvalue[1].replaceAll("%", "");
			keyvalue[1] = keyvalue[1].replaceAll("\\*", "");
		}
		myQuery += "WHERE ";
		// example: select id from media where key='artist' and value='EAV'
		for(int i=0; i < keyvalue.length-1; i+=2) {
			myQuery += "key" + equ + "'" + keyvalue[i] + "' AND value" + equ + "'" + keyvalue[i+1] + "'";
		}
		return myQuery;
	}

	public ReturnValue mlibGetPlaylists() {
		String myQuery = "SELECT name, id FROM Playlist ORDER BY id;";
		logger.debug(myQuery);
		result = Xmmsclient.xmmsc_medialib_select(myConnection, myQuery);
		Xmmsclient.xmmsc_result_wait(result);
		if (Xmmsclient.xmmsc_result_is_list(result) == 1){
			Xmmsclient.xmmsc_result_list_first(result);
			int id[] = new int[1];
			String name[] = new String[1];
			name[0] = "";
			List entries = new LinkedList();
			Map ids = new HashMap();
			while (Xmmsclient.xmmsc_result_list_valid(result) == 1){
				Xmmsclient.xmmsc_result_get_dict_entry_int32(result,"id", id);
				Xmmsclient.xmmsc_result_get_dict_entry_str(result,"name", name);
				Xmmsclient.xmmsc_result_list_next(result);
				Integer tmpint = new Integer(id[0]);
				if(!ids.containsKey(tmpint) && !name[0].startsWith("_"))
					ids.put(new Integer(id[0]),name[0]);
			}
			Set tmpIds = ids.keySet();
			for (Iterator i = tmpIds.iterator(); i.hasNext();){
				Object current = i.next();
				PlaylistInformation tmp = getPlaylistInformationForID(((Integer)current).intValue());
				if (tmp != null) {
					tmp.setName((String)ids.get(current));
					entries.add(tmp);
				}
			}
			ret.setType(ReturnValue.PLAYLISTINFORMATION_TYPE);
			ret.setValue(entries);
			return ret;
		}
		ret.setType(ReturnValue.ERROR_TYPE);
		return ret;
	}
	
	public ReturnValue mlibSelect(String[] url, boolean search) {
		return mlibSelect(buildQuery(url), search);
	}

	public ReturnValue mlibSelect(String query, boolean search) {
		logger.debug(query);
		result = Xmmsclient.xmmsc_medialib_select(myConnection, query);
		Xmmsclient.xmmsc_result_wait(result);
		if (Xmmsclient.xmmsc_result_is_list(result) == 1){
			Xmmsclient.xmmsc_result_list_first(result);
			int id[] = new int[1];
			List entries = new LinkedList();
			Set ids = new HashSet();
			while (Xmmsclient.xmmsc_result_list_valid(result) == 1){
				Xmmsclient.xmmsc_result_get_dict_entry_int32(result,"id", id);
				Xmmsclient.xmmsc_result_list_next(result);
				Integer tmpint = new Integer(id[0]);
				if(!ids.contains(tmpint))
					ids.add(new Integer(id[0]));
			}
			for (Iterator i = ids.iterator(); i.hasNext();){
				int nid = ((Integer)i.next()).intValue();
				SWIGTYPE_p_xmmsc_result_St res = Xmmsclient.xmmsc_medialib_get_info(myConnection, nid);
				Title tmp = getTitleForID(res, nid);
				Xmmsclient.xmmsc_result_unref(res);
				if (tmp != null)
					entries.add(tmp);
			}
			if (search)
				ret.setType(ReturnValue.MEDIALIB_SEARCH_TYPE);
			else
				ret.setType(ReturnValue.MEDIALIB_TYPE);
			ret.setValue(entries);
			return ret;
		}
		ret.setType(ReturnValue.ERROR_TYPE);
		return ret;
	}

	public ReturnValue mlibLoadPlaylist(String name) {
		result = Xmmsclient.xmmsc_playlist_clear(myConnection);
		Xmmsclient.xmmsc_result_wait(result);
		return mlibAddPlaylistToCurrent(name);
	}
	
	public ReturnValue mlibLoadPlaylist(String url, String name) {
		result = Xmmsclient.xmmsc_medialib_playlist_import(myConnection, name, url);
		Xmmsclient.xmmsc_result_wait(result);
		// TODO workaround until xmms2 team gets a mlib_playlist_added broadcast
		setChanged();
		notifyObservers(new UpdateUIType(UpdateUIType.PLAYLIST_LOADED, null));
		return ret;
	}
	
	public ReturnValue mlibAddPlaylistToCurrent(String name) {
		result = Xmmsclient.xmmsc_medialib_playlist_load(myConnection, name);
		Xmmsclient.xmmsc_result_wait(result);
		return ret;		
	}
	
	public ReturnValue mlibDeletePlaylist(String name) {
		result = Xmmsclient.xmmsc_medialib_playlist_remove(myConnection, name);
		Xmmsclient.xmmsc_result_wait(result);
		return ret;	
	}

	public ReturnValue mlibRenamePlaylist(String nameold, String namenew) {
		this.mlibLoadPlaylist(nameold);
		
		result = Xmmsclient.xmmsc_medialib_playlist_save_current(myConnection, namenew);
		Xmmsclient.xmmsc_result_wait(result);
	
		return mlibDeletePlaylist(nameold);	
	}

	public ReturnValue mlibNewPlaylist(String name) {
		result = Xmmsclient.xmmsc_playlist_clear(myConnection);
		Xmmsclient.xmmsc_result_wait(result);
		result = Xmmsclient.xmmsc_medialib_playlist_save_current(myConnection, name);
		Xmmsclient.xmmsc_result_wait(result);
		return ret;	
	}

	public ReturnValue mlibSaveCurrentPlaylist(String name) {
		result = Xmmsclient.xmmsc_medialib_playlist_save_current(myConnection, name);
		Xmmsclient.xmmsc_result_wait(result);
		return ret;	
	}

	public ReturnValue mlibGetTitleForID(int id) {
		result = Xmmsclient.xmmsc_medialib_get_info(myConnection, id);
		ret.setType(ReturnValue.TITLE_TYPE);
		ret.setValue(getTitleForID(result, id));
		return ret;
	}

	public ReturnValue mlibUpdateTitle(Title t) {
		/*if (t.getAttribute("url").toLowerCase().endsWith(".mp3") || t.getAttribute("url").toLowerCase().endsWith(".ogg")){
			try {
				updateMP3(t);
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (TagException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			result = Xmmsclient.xmmsc_medialib_entry_property_set(myConnection, 
					Integer.parseInt(t.getAttribute("id")), "url", t.getAttribute("url"));
			Xmmsclient.xmmsc_result_wait(result);
			return mlibRefresh(Integer.parseInt(t.getAttribute("id")));
		}*/
		result = Xmmsclient.xmmsc_medialib_entry_property_set_with_source(myConnection, 
				Integer.parseInt(t.getAttribute("id")), Xmms2ThreadFactory.mySource, "url", t.getAttribute("url"));
		Xmmsclient.xmmsc_result_wait(result);
		if (isError(result)) return ret;
		Xmmsclient.xmmsc_result_unref(result);
		result = Xmmsclient.xmmsc_medialib_entry_property_set_with_source(myConnection, 
				Integer.parseInt(t.getAttribute("id")), Xmms2ThreadFactory.mySource, "artist", t.getAttribute("artist"));
		Xmmsclient.xmmsc_result_wait(result);
		if (isError(result)) return ret;
		Xmmsclient.xmmsc_result_unref(result);
		result = Xmmsclient.xmmsc_medialib_entry_property_set_with_source(myConnection, 
				Integer.parseInt(t.getAttribute("id")), Xmms2ThreadFactory.mySource, "album", t.getAttribute("album"));
		Xmmsclient.xmmsc_result_wait(result);
		if (isError(result)) return ret;
		Xmmsclient.xmmsc_result_unref(result);
		result = Xmmsclient.xmmsc_medialib_entry_property_set_with_source(myConnection, 
				Integer.parseInt(t.getAttribute("id")), Xmms2ThreadFactory.mySource, "title", t.getAttribute("title"));
		Xmmsclient.xmmsc_result_wait(result);
		if (isError(result)) return ret;
		Xmmsclient.xmmsc_result_unref(result);
		result = Xmmsclient.xmmsc_medialib_entry_property_set_with_source(myConnection, 
				Integer.parseInt(t.getAttribute("id")), Xmms2ThreadFactory.mySource, "comment", t.getAttribute("comment"));
		Xmmsclient.xmmsc_result_wait(result);
		mm.updateTitleCache(Integer.parseInt(t.getAttribute("id")));
		return ret;
	}
	
	public ReturnValue mlibRefresh() {
		return mlibRefresh(0);
	}
	
	/**
	 * Refresh some specific medialib id
	 * @param id
	 * @return		empty ReturnValue
	 */
	private ReturnValue mlibRefresh(int id){
		result = Xmmsclient.xmmsc_medialib_rehash(myConnection, id);
		Xmmsclient.xmmsc_result_wait(result);
		return ret;
	}
	
	/**
	 * not used atm
	 * @param t
	 * @throws IOException
	 * @throws TagException
	 */
	private void updateMP3(Title t) throws IOException, TagException{
		//TODO doesn't work atm
		System.out.println(t.getAttribute("url"));
		File f = null;
		try {
			f = new File(URLDecoder.decode(t.getAttribute("url"), "UTF-8"));
		} catch (UnsupportedEncodingException e) {
			logger.debug(e.getMessage(), e);
		} 
		if (f.exists()){
			logger.debug("gonna update " + f);
			MP3File mp3 = new MP3File(f);
			ID3v1 tag = mp3.getID3v1Tag();
			if (tag != null){
				System.out.println("updating " + mp3);
				tag.setAlbum(t.getAttribute("album"));
				tag.setArtist(t.getAttribute("artist"));
				tag.setComment(t.getAttribute("comment"));
				tag.setTitle(t.getAttribute("title"));
				mp3.setID3v1Tag(tag);
				mp3.save();
			}
		}
	}
	
	protected String spinDown(){
		String error = null;
		if (result == null)
			error = "calling the requested function didn't succeed";
		else if (Xmmsclient.xmmsc_result_iserror(result) == 1)
			error = Xmmsclient.xmmsc_result_get_error(result);
		if (result != null) Xmmsclient.xmmsc_result_unref(result);
		synchronized (Xmms2ThreadFactory.lockMlibConnection){
			Xmms2ThreadFactory.lockMlibConnection.notify();
			Xmms2ThreadFactory.lockMlibConnection = Boolean.FALSE;
		}
		return error;
	}
	
	protected boolean enterCritical(){
		synchronized(Xmms2ThreadFactory.lockMlibConnection){
			while (Xmms2ThreadFactory.lockMlibConnection.equals(Boolean.TRUE)){
				try {
					Xmms2ThreadFactory.lockMlibConnection.wait();
				} catch (InterruptedException e) {
					logger.debug(e.getMessage(), e);
					return false;
				}
			}
			Xmms2ThreadFactory.lockMlibConnection = Boolean.TRUE;
			return true;
		}
	}
}
