/******************************************************************
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ******************************************************************/

package core.xmms2;

import java.util.Observer;

import org.xmms2.xmms2bindings.*;

import statics.StaticValues;

import core.KeepStateThread;
import core.ManagedThread;
import core.Management;
import core.ReturnValue;
import core.ThreadFactory;
import core.UpdateUIType;

/**
 * Xmms2 Threadfactory. You get one KeepStateThread, millions of ControlThread and almost the same 
 * #MedialibThreads ;). For description of reused methods have a look at core.ThreadFactory
 */
public final class Xmms2ThreadFactory extends ThreadFactory{
	private SWIGTYPE_p_xmmsc_connection_St controlConnection;
	private SWIGTYPE_p_xmmsc_connection_St mlibConnection;
	private SWIGTYPE_p_xmmsc_connection_St callbackConnection;
	private KeepStateThread kst;
	private boolean spinDown = false;
	protected static String[] sourcePrefence = new String[]{"clients/X4X", 
		"clients/generic", "clients/*", "server", "core/*", "plugins/*"};
	protected static String mySource = "clients/X4X";
	protected static Boolean lockControlConnection = Boolean.FALSE;
	protected static Boolean lockMlibConnection = Boolean.FALSE;
	
	/**
	 * Creates a new Xmms2ThreadFactory which means: load the library and init the connections to the daemon 
	 * 
	 * @param mm 
	 * @throws Exception 
	 *
	 */
	public Xmms2ThreadFactory(Management mm) throws Exception{
		super(mm);
		try {
			System.loadLibrary("xmms2java");
		} catch (UnsatisfiedLinkError e){
			logger.fatal(StaticValues.getLangString("clientlibUnavailable"));
			logger.debug(e.getMessage(), e);
			throw new Exception(StaticValues.getLangString("clientlibUnavailable"));
		} catch (Throwable t){
			logger.debug(t.getMessage(), t);
		}
		
		defaultSQLString = Xmms2MedialibThread.defaultQuery;
		controlConnection = Xmmsclient.xmmsc_init("X4X-Control");
		mlibConnection = Xmmsclient.xmmsc_init("X4X-Medialib");
		callbackConnection = Xmmsclient.xmmsc_init("X4X-Callbacks");
		connect();
	}
	
	public ManagedThread getControlThread(ReturnValue ret, Observer obs) {
		if (!connected) connect();
		if (connected) return new Xmms2ControlThread(ret, obs, mm, controlConnection);
		return null;
	}

	public ManagedThread getMedialibThread(ReturnValue ret, Observer obs) {
		if (!connected) connect();
		if (connected) return new Xmms2MedialibThread(ret, obs, mm, mlibConnection);
		return null;
	}
	
	public KeepStateThread getKeepStateThread(Observer obs){
		if(kst == null)
			kst = new Xmms2KeepStateThread(this, obs, mm, callbackConnection);
		return kst;
	}
	
	public void spinDown() {
		logger.info("Spinning down");
		spinDown = true;
		if (connected){
			Xmmsclient.xmmsc_disconnect_callback_set(controlConnection, null, new int[]{0});
			Xmmsclient.xmmsc_disconnect_callback_set(mlibConnection, null, new int[]{0});
			Xmmsclient.xmmsc_disconnect_callback_set(callbackConnection, null, new int[]{0});
		}
		((Xmms2KeepStateThread)kst).spinDown();
		
		Xmmsclient.xmmsc_unref(controlConnection);
		Xmmsclient.xmmsc_unref(mlibConnection);
		Xmmsclient.xmmsc_unref(callbackConnection);
	}
	
	/**
	 * Connects a specific connection to xmms2d
	 * @param conn
	 */
	private void connect(SWIGTYPE_p_xmmsc_connection_St conn){
		if (!spinDown){
			String path = null;
			if (StaticValues.getSetting("serverip").equals("") && StaticValues.getSetting("serverport").equals(""))
				path = System.getProperty("XMMS_PATH");
			else path = "tcp://" + StaticValues.getSetting("serverip") + ":" + StaticValues.getSetting("serverport");
			if (path != null && (path.equals("") || path.equals("${env.XMMS_PATH}"))) path = null;
			if (Xmmsclient.xmmsc_connect(conn, path) <= 0){
				logger.warn(StaticValues.getLangString("xmms2dUnavailable"));
				connected = false;
			}
			else {
				connected = true;
				Xmmsclient.xmmsc_disconnect_callback_set(conn, XmmsclientConstants.DISCONNECT_CALLBACK, new int[]{0});
			}
		}
	}
	
	public void connect(){
		connect(controlConnection);
		connect(mlibConnection);
		connect(callbackConnection);
		if (connected)
			mm.restartBackend();
		else {
			setChanged();
			notifyObservers(new UpdateUIType(UpdateUIType.BACKEND_LOST, new Object[]{}));
		}
	}
}
