/******************************************************************
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Copyright:
 * Georg Schild (dangertools@gmail.com)
 * Franz Endstrasser (franz.endstrasser@gmail.com)
 ******************************************************************/

package gui;

import java.lang.reflect.Field;
import java.util.ListIterator;

import org.Xwidgets.Styles;
import org.Xwidgets.WidgetFactory;
import org.Xwidgets.XButton;
import org.Xwidgets.XLCDDisplay;
import org.Xwidgets.XPanel;
import org.apache.log4j.Logger;
import org.jdom.DataConversionException;
import org.jdom.Element;

import statics.StaticValues;

/**
 * That class simply represents <u>one</u>, and only one ButtonBar. It is, as every other
 * "ui-part class" in that package a singleton. The ButtonBar is dynamically built from
 * a jdom-Element and the Buttons are instantiated using some Reflection. You usually want
 * to use the Field me for usage in another XPanel or XFrame
 */

public class ButtonBar {
	private static ButtonBar instance = null;
	protected XPanel me;
	private Logger logger;
	protected XLCDDisplay lcd;
	
	protected XButton play = null;
	protected XButton stop = null;
	protected XButton pause = null;
	protected XButton next = null;
	protected XButton prev = null;
	protected XButton about = null;
	protected XButton lists = null;
	
	/**
	 * Create ButtonBar Object. used only once cause it's singleton
	 * 
	 * @param parent		parent Element (maybe dummy)
	 * @param bar			jdom Element
	 */
	private ButtonBar(XPanel parent, Element bar){
		logger = Logger.getLogger(getClass());
		try {
			buildMe(parent, bar);
		}catch (Exception e){
			logger.fatal(StaticValues.getLangString("uiError"));
			logger.debug(e.getMessage(), e);
			
		}
	}

	/**
	 * The method returns a singleton ButtonBar
	 * 
	 * @param parent		the ButtonBar's parentpanel
	 * @param bar			the jdom Element represent the bar
	 * @return				returns a singleton Object of a ButtonBar
	 */
	
	protected static ButtonBar getButtonBar(XPanel parent, Element bar){
		if (instance == null)
			instance = new ButtonBar(parent, bar);
		return instance;
	}
	
	/**
	 * Called by Constructor to build everything
	 * 
	 * @param parent
	 * @param bar
	 */
	private void buildMe(XPanel parent, Element bar){
		WidgetFactory wt = MainGUI.mainGui.myGUIFactory;
		int size = 0;
		String tmp = bar.getAttributeValue("size");
		if (tmp.equals("small")) size = 24;
		else if (tmp.equals("normal")) size = 40;
		else if (tmp.equals("medium")) size = 56;
		else if (tmp.equals("big")) size = 72;
		
		//if (bar.getAttributeValue("direction").equals("horizontal"))
			me = wt.getPanel(parent, parent.getWidth(), size+2, Styles.SIMPLE);
		//else panel = wt.getPanel(parent, size+2, parent.getHeight(), Styles.BORDER);
		int last = 0;
		for ( ListIterator it = bar.getChildren().listIterator(); it.hasNext(); ){
			Element child = (Element)it.next();
			if (!child.getName().endsWith("Button"))logger.info(StaticValues.getLangString("unknownElement"));
			else {
				String fieldname = "";
				try {
					fieldname = child.getName().toLowerCase().replaceAll("button", "");
					Field but = getClass().getDeclaredField(fieldname);
					last = it.nextIndex()-1;
					but.set(this, getButton(wt, child, size, me, bar.getAttributeValue("direction"), last));
				} catch (SecurityException e) {
					logger.debug(e.getMessage(), e);
				} catch (NoSuchFieldException e) {
					logger.info(StaticValues.getLangString("ReflectionFieldException") + " \"" + 
							fieldname + "\": " + StaticValues.getLangString("noSuchField"));
					logger.debug(e.getMessage(), e);
				} catch (IllegalArgumentException e) {
					logger.debug(e.getMessage(), e);
				} catch (IllegalAccessException e) {
					logger.info(StaticValues.getLangString("ReflectionMethodException") + " \"" + 
							fieldname + "\": " + StaticValues.getLangString("noAccess"));
					logger.debug(e.getMessage(), e);
				} 
			}
		}
		
		lcd = wt.getLCDDisplay(me, (int)(size*3.0), size, Styles.SIMPLE);
		lcd.setNumberOfDigits(2);
		lcd.setLocation((last+2)*size, 0);
		
		try {
			if (bar.getAttribute("packed").getBooleanValue() == true)
				me.pack();
		} catch (DataConversionException e) {
			logger.debug(e.getMessage(), e);
		}
	}

	/**
	 * Builds a button on demand. called from buildMe()
	 * 
	 * @param wt
	 * @param button
	 * @param size	
	 * @param panel
	 * @param direction
	 * @param index
	 * @return				built XButton
	 */
	private XButton getButton(WidgetFactory wt, Element button, int size, XPanel panel, String direction, int index) {
		XButton b = null;
		try {
			b = wt.getButton(panel, size, size, StaticValues.getLangString(button.getName()), Styles.SIMPLE);
			b.setImage(getImageForButton(button, size));
			b.setTooltipText(StaticValues.getLangString(button.getName()));
			//if (direction.equals("horizontal"))
				b.setLocation(size*(index), 0);
			//else
				//b.setLocation(0, size*(index));
			b.addMouseListener(MouseListener.getInstance());
		}catch (Exception e) {
			logger.debug(e.getMessage(), e);
		}
		return b;
	}
	
	/**
	 * Get Image for Button
	 * 
	 * @param button
	 * @param size
	 * @return			String to right Image
	 */
	private String getImageForButton(Element button, int size) {
		String prefix = MainGUI.mainGui.iconPrefix + button.getName().toLowerCase().replaceAll("button", "");
		switch (size){
			case 24:	return prefix + "16x16" + ".png";
			case 40:	return prefix + "32x32" + ".png";
			case 56:	return prefix + "48x48" + ".png";
			case 72:	return prefix + "64x64" + ".png";
			default:	return "";
		}
	}
}
