/******************************************************************
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * 
 * Copyright:
 * Georg Schild (dangertools@gmail.com)
 * Franz Endstrasser (franz.endstrasser@gmail.com)
 ******************************************************************/

package gui;

import java.awt.Toolkit;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.channels.FileChannel;
import java.util.Calendar;
import java.util.List;
import java.util.ListIterator;
import java.util.Locale;

import javax.swing.JOptionPane;

import gui.models.DNA;

import org.Xwidgets.Styles;
import org.Xwidgets.WidgetFactory;
import org.Xwidgets.XFlowText;
import org.Xwidgets.XFrame;
import org.Xwidgets.XLabel;
import org.Xwidgets.XMenu;
import org.Xwidgets.XMenuBar;
import org.Xwidgets.XMenuItem;
import org.Xwidgets.XMessageBox;
import org.Xwidgets.XPanel;
import org.Xwidgets.XSlider;
import org.Xwidgets.XSplitPane;
import org.Xwidgets.XSystemTray;
import org.Xwidgets.XTabbedPane;
import org.Xwidgets.XWidgetsException;
import org.Xwidgets.SWT.SWTFactory;
import org.apache.log4j.Logger;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.jdom.DataConversionException;
import org.jdom.Element;

import core.Management;

import statics.StaticValues;

/**
 * Main class of X4X. This is the only class with a main() Method. It just builds and defines the UI-Elements,
 * some things at startup are done too. No computation happens here
 */
public final class MainGUI {
	protected ButtonBar buttonPanel;
	protected PlaylistPanel playlistPanel;
	protected EQPanel eqPanel;
	protected NetworkBrowser CIFSBrowser;
	protected XSlider volume, playtime;
	protected XTabbedPane tabbedPane;
	private Logger logger;
	protected XFrame myFrame, lists, eqWindow;
	protected WidgetFactory myGUIFactory;
	private DNA dna;
	protected XFlowText flow;
	protected PropGUI propGui;
	protected PrefGUI prefGui;
	protected MlibPanel mlibPanel;
	protected XMenuItem menuExit, menuPrefs, menuMlib, aboutEntry, helpEntry, eqEntry, 
			trayExit, trayAbout, trayPlay, trayStop, trayPause, trayNext, trayPrev;
	protected Object layoutWithBorders, layoutDataFillHoriz, layoutDataFillBoth, 
			layoutWithoutBorders, headerLayout;
	protected boolean tabbed, external;
	protected XLabel status;
	private Element me;
	protected XSystemTray tray;
	public static final String version = "0.1.2-Grete";
	protected String iconPrefix;
	
	protected static MainGUI mainGui;
	
	/**
	 * construct and start the mainGUI
	 */
	private MainGUI(String[] args){
		initialize();
		iconPrefix = StaticValues.getSetting("iconDir") + StaticValues.getSetting("iconstyle") + StaticValues.getSetting("deli");
		mainGui = this;
		logger = Logger.getLogger(getClass());
		logger.info("==== " + Calendar.getInstance(Locale.getDefault()).getTime() + " ====");
		logger.info("X4X version " + version + " started");
		logger.info("If you need more output open log4j.properties and set rootLogger's value to debug");
		dna = new DNA();
		Object mainGUILayout = null;
		if (StaticValues.getSetting("gui").equals("swt")){
			try {
				myGUIFactory = SWTFactory.getInstance();
			} catch (Throwable e) {
				JOptionPane.showMessageDialog(null, "Unable to find SWT Library.\nPlease "+
						"(re-)install SWT", "SWT missing", JOptionPane.ERROR_MESSAGE);
				System.exit(1);
			}
			layoutWithBorders = new GridLayout(1, false);
			((GridLayout)layoutWithBorders).marginWidth = 30;
			((GridLayout)layoutWithBorders).marginHeight = 50;
			((GridLayout)layoutWithBorders).verticalSpacing = 10;
			layoutWithoutBorders = new GridLayout(1, false);
			((GridLayout)layoutWithoutBorders).verticalSpacing = 10;
			
			layoutDataFillHoriz = new GridData(GridData.FILL_HORIZONTAL);
			((GridData)layoutDataFillHoriz).grabExcessHorizontalSpace = true;
			((GridData)layoutDataFillHoriz).heightHint = 27;
			
			layoutDataFillBoth = new GridData(GridData.FILL_BOTH);
			((GridData)layoutDataFillBoth).grabExcessVerticalSpace = true;
			((GridData)layoutDataFillBoth).grabExcessHorizontalSpace = true;
			
			mainGUILayout = new GridLayout(1, false);
			((GridLayout)mainGUILayout).verticalSpacing = 5;
			((GridLayout)mainGUILayout).marginWidth = 30;
		}
		myGUIFactory.setAppName("X4X");
		try {
			me = dna.getBaseElement("MainGUI");
			if (me == null){
				logger.fatal(StaticValues.getLangString("noMe"));
				XMessageBox error = myGUIFactory.getMessageBox(null, Styles.SIMPLE);
				error.setText(StaticValues.getLangString("noMe"));
				error.open();
				System.exit(1);
			}
			if (me.getAttribute("resizeable").getBooleanValue())
				myFrame = myGUIFactory.getFrame(null, -1, -1, Styles.RESIZE);
			else myFrame = myGUIFactory.getFrame(null, -1, -1, Styles.SIMPLE);
			myFrame.setLayout(mainGUILayout);
			
			myFrame.setImage(iconPrefix + "trayIcon.png");
			
			List masters = dna.getChildren(me);
			for (ListIterator it = masters.listIterator(); it.hasNext();){
				Element current = (Element)it.next();
				if (current.getName().equals("currentTitle")){
					flow = myGUIFactory.getFlowText(myFrame, 400, 30, Styles.RIGHT);
					flow.setSpeed(150);
					flow.setLayoutData(layoutDataFillHoriz);
				}
				else if (current.getName().equals("VolumeSlider")){
					volume = buildSlider(current);
					volume.setLayoutData(layoutDataFillHoriz);
					volume.setMax(100);
				}
				else if (current.getName().equals("PlaytimeSlider")){
					playtime = buildSlider(current);
					playtime.setLayoutData(layoutDataFillHoriz);
				}
				else if (current.getName().equals("Lists")){
					buildPanels(dna.getChild(me, "Lists"));
				}
				else if (current.getName().equals("ButtonBar")){
					buttonPanel = ButtonBar.getButtonBar(myFrame, current);
				}
				else {
					logger.debug(StaticValues.getLangString("unknownElement") + ": " + current.getName());
					XMessageBox error = myGUIFactory.getMessageBox(myFrame, Styles.OK|Styles.ERROR);
					error.setText(StaticValues.getLangString("unknownElement"));
					error.open();
				}
			}
			
			status = myGUIFactory.getLabel(myFrame, -1, -1, "", Styles.LEFT);
			status.setLayoutData(layoutDataFillHoriz);
			
			CIFSBrowser = new NetworkBrowser(layoutWithBorders, layoutDataFillHoriz, myGUIFactory);
			
			buildMenu();
			buildTray();
			
			myFrame.setText("X4X @" + StaticValues.getSetting("backend").toUpperCase());

			new UpdateListener(args);
			propGui = new PropGUI();
			prefGui = new PrefGUI();
		} catch (Throwable e){
			logger.debug(e.getMessage(), e);
		}
	}
	
	private void open(){
		try {
			if (!external && tabbedPane != null){
				myFrame.setSize(me.getAttribute("width").getIntValue(), 
						me.getAttribute("height").getIntValue());
			}
			else myFrame.pack();
			if (!external && tabbedPane == null)
				myFrame.setSize(Toolkit.getDefaultToolkit().getScreenSize().width-5,
						Toolkit.getDefaultToolkit().getScreenSize().height-5);
			myGUIFactory.run(this, getClass().getDeclaredMethod("finalizer", new Class[]{}), new Object[]{});
		} catch (DataConversionException e) {
			logger.debug(e.getMessage(), e);
		} catch (SecurityException e) {
			logger.debug(e.getMessage(), e);
		} catch (NoSuchMethodException e) {
			logger.debug(e.getMessage(), e);
		}
	}
	
	private void buildTray(){		
		tray = myGUIFactory.getSystemTray(Styles.NULL);
		tray.setName("X4X");
		tray.setTooltipText("X4X");
		tray.setImage(iconPrefix + "trayIcon.png");
		
		XMenu trayMenu = myGUIFactory.getDropDownMenu(myFrame);
		
		trayPlay = myGUIFactory.getMenuItem(trayMenu);
		trayPlay.setText(StaticValues.getLangString("PlayButton"));
		trayPlay.addSelectionListener(SelectionListener.getInstance());
		trayPlay.setImage(iconPrefix + "trayPlay.png");
		
		trayPause = myGUIFactory.getMenuItem(trayMenu);
		trayPause.setText(StaticValues.getLangString("PauseButton"));
		trayPause.addSelectionListener(SelectionListener.getInstance());
		trayPause.setImage(iconPrefix + "trayPause.png");
		
		trayStop = myGUIFactory.getMenuItem(trayMenu);
		trayStop.setText(StaticValues.getLangString("StopButton"));
		trayStop.addSelectionListener(SelectionListener.getInstance());
		trayStop.setImage(iconPrefix + "trayStop.png");
		
		trayNext = myGUIFactory.getMenuItem(trayMenu);
		trayNext.setText(StaticValues.getLangString("NextButton"));
		trayNext.addSelectionListener(SelectionListener.getInstance());
		trayNext.setImage(iconPrefix + "trayNext.png");
		
		trayPrev = myGUIFactory.getMenuItem(trayMenu);
		trayPrev.setText(StaticValues.getLangString("PrevButton"));
		trayPrev.addSelectionListener(SelectionListener.getInstance());
		trayPrev.setImage(iconPrefix + "trayPrev.png");
		
		myGUIFactory.getMenuSeparator(trayMenu);
		
		trayAbout = myGUIFactory.getMenuItem(trayMenu);
		trayAbout.setText(StaticValues.getLangString("aboutEntry"));
		trayAbout.addSelectionListener(SelectionListener.getInstance());
		trayAbout.setImage(iconPrefix + "trayAbout.png");
		
		trayExit = myGUIFactory.getMenuItem(trayMenu);
		trayExit.setText(StaticValues.getLangString("exitEntry"));
		trayExit.addSelectionListener(SelectionListener.getInstance());
		trayExit.setImage(iconPrefix + "trayExit.png");
		
		tray.setMenu(trayMenu);
	}

	/**
	 * Build menu
	 */
	private void buildMenu() {
		XMenuBar mbar = myGUIFactory.getMenuBar(myFrame, new String[]{StaticValues.getLangString("fileMenu"), 
				StaticValues.getLangString("helpMenu")});
		if (external){
			menuMlib = myGUIFactory.getMenuItem(mbar.getMenuAt(0));
			menuMlib.setText(StaticValues.getLangString("lists"));
			menuMlib.addSelectionListener(SelectionListener.getInstance());
			menuMlib.setImage(iconPrefix + "trayLists.png");
		}
		
		menuPrefs = myGUIFactory.getMenuItem(mbar.getMenuAt(0));
		menuPrefs.setText(StaticValues.getLangString("preferences"));
		menuPrefs.addSelectionListener(SelectionListener.getInstance());
		menuPrefs.setImage(iconPrefix + "trayPrefs.png");
		
		eqEntry = myGUIFactory.getMenuItem(mbar.getMenuAt(0));
		eqEntry.setText(StaticValues.getLangString("eqEntry"));
		eqEntry.addSelectionListener(SelectionListener.getInstance());
		eqEntry.setImage(iconPrefix + "trayEQ.png");
		
		myGUIFactory.getMenuSeparator(mbar.getMenuAt(0));
		
		menuExit = myGUIFactory.getMenuItem(mbar.getMenuAt(0));
		menuExit.setText(StaticValues.getLangString("exitEntry"));
		menuExit.addSelectionListener(SelectionListener.getInstance());
		menuExit.setImage(iconPrefix + "trayExit.png");
		
		helpEntry = myGUIFactory.getMenuItem(mbar.getMenuAt(1));
		helpEntry.setText(StaticValues.getLangString("helpEntry"));
		helpEntry.addSelectionListener(SelectionListener.getInstance());
		helpEntry.setImage(iconPrefix + "trayHelp.png");
		
		myGUIFactory.getMenuSeparator(mbar.getMenuAt(1));
		
		aboutEntry = myGUIFactory.getMenuItem(mbar.getMenuAt(1));
		aboutEntry.setText(StaticValues.getLangString("aboutEntry"));
		aboutEntry.addSelectionListener(SelectionListener.getInstance());
		aboutEntry.setImage(iconPrefix + "trayAbout.png");
		
		myFrame.setMenuBar(mbar);
	}
	
	/**
	 * Build slider from Element
	 * 
	 * @param el		jdom Element
	 * @return			XSlider
	 */
	private XSlider buildSlider(Element el) {
		XSlider s = myGUIFactory.getSlider(myFrame, -1, -1, Styles.HORIZONTAL);
		s.addMouseListener(MouseListener.getInstance());	
		return s;
	}
	
	/**
	 * build panels (playlist/playlists/mlib-search/mlib-browse)
	 * 
	 * @param dnaLists		jdom Element describing look of lists
	 */
	private void buildPanels(Element dnaLists) {
		tabbed = true;
		external = true;
		try {
			tabbed = dnaLists.getAttribute("tabbed").getBooleanValue();
			external = dnaLists.getAttribute("external").getBooleanValue();
		} catch (DataConversionException e1) {
			logger.debug(e1.getMessage(), e1);
		}
		try {
			XPanel dummy = myFrame;
			if (external){
				lists = myGUIFactory.getFrame(myFrame, 600, 600, Styles.RESIZE);
				lists.closeOnClose(false);
				lists.setImage(iconPrefix + "trayLists.png");
				lists.setText(StaticValues.getLangString("lists"));
				lists.setLayout(layoutWithoutBorders);
				lists.setLayoutData(layoutDataFillHoriz);
				dummy = lists;
			}
			if(tabbed) {
				playlistPanel = PlaylistPanel.getPlaylistPanel(dummy, dna.getChild(dnaLists, "PlaylistPanel"), 
						layoutWithoutBorders, layoutDataFillHoriz);
				mlibPanel = MlibPanel.getInstance(dummy, dna.getChild(dnaLists, "MedialibPanel"), 
						layoutWithoutBorders, layoutDataFillHoriz);
				tabbedPane = myGUIFactory.getTabbedPane(dummy, -1, -1, Styles.SIMPLE);
				tabbedPane.setLayout(layoutWithoutBorders);
				tabbedPane.setLayoutData(layoutDataFillBoth);
				tabbedPane.addSelectionListener(SelectionListener.getInstance());
				tabbedPane.addItem(playlistPanel.panel,StaticValues.getLangString("playlist"), 
						StaticValues.getLangString("playlistHeader"));
				tabbedPane.addItem(playlistPanel.playlistsPanel,StaticValues.getLangString("playlists"),
						StaticValues.getLangString("playlistsHeader"));
				tabbedPane.addItem(mlibPanel.browsePanel, StaticValues.getLangString("mlibBrowse"), 
						StaticValues.getLangString("mlibBrowseHeader"));
				tabbedPane.addItem(mlibPanel.searchPanel, StaticValues.getLangString("mlibSearch"), 
						StaticValues.getLangString("mlibSearchHeader"));
			}
			else {
				XSplitPane mlibSplit, playlistSplit;
				playlistPanel = PlaylistPanel.getPlaylistPanel(dummy, dna.getChild(dnaLists, "PlaylistPanel"), 
						layoutWithoutBorders, layoutDataFillHoriz);
				mlibPanel = MlibPanel.getInstance(dummy, dna.getChild(dnaLists, "MedialibPanel"), 
						layoutWithoutBorders, layoutDataFillHoriz);
				XSplitPane splitPane = myGUIFactory.getSplitPane(dummy, -1, -1, Styles.VERTICAL);
				splitPane.setLayout(layoutWithoutBorders);
				splitPane.setLayoutData(layoutDataFillBoth);
				splitPane.setItems((mlibSplit = myGUIFactory.getSplitPane(splitPane, -1, -1, Styles.HORIZONTAL)), 
						(playlistSplit = myGUIFactory.getSplitPane(splitPane, -1, -1, Styles.HORIZONTAL)));
				mlibSplit.setLayout(layoutWithoutBorders);
				mlibSplit.setLayoutData(layoutDataFillBoth);
				playlistSplit.setLayout(layoutWithoutBorders);
				playlistSplit.setLayoutData(layoutDataFillBoth);
				mlibSplit.setItems(mlibPanel.browsePanel, mlibPanel.searchPanel);
				playlistSplit.setItems(playlistPanel.playlistsPanel, playlistPanel.panel);
			}
		} catch (XWidgetsException e) {
			logger.debug(e.getMessage(), e);
		}
	}
	
	/**
	 * Only method which is used elsewhere too. A listener could call it got get a really actual EQPanel
	 *
	 */
	protected void buildEQ(){
		int val = 0;
		while(Management.getInstance().getEQ(val) > -51)
			val++;
		if (eqWindow == null || eqPanel.eqCount != val){
			eqWindow = myGUIFactory.getFrame(myFrame, 400, 150, Styles.SIMPLE);
			eqWindow.setText(StaticValues.getLangString("eq"));
			eqWindow.closeOnClose(false);
			eqWindow.setImage(iconPrefix + "trayEQ.png");
			try {
				eqWindow.setLayout(layoutWithoutBorders);
			} catch (XWidgetsException e) {
				logger.debug(e.getMessage(), e);
			}
			eqPanel = EQPanel.getEQPanel(eqWindow, null, val);
			eqWindow.pack();
		}
	}
	
	/**
	 * Called on finalization from the WidgetFactory
	 *
	 */
	public void finalizer(){
		StaticValues.finish();
		Management.getInstance().spinDown();
		System.exit(0);
	}
	
	/**
	 * Called on first startup to copy dirs etc.
	 */
	private void initialize() {
		File confDir = new File(StaticValues.getSetting("prefdir"));
		if (!confDir.exists()){
			Logger.getLogger(MainGUI.class).debug("Creating preferences directory at " + confDir);
			copyDir(new File(StaticValues.getSetting("basedir") + 
					StaticValues.getSetting("deli") + ".xmms2" + StaticValues.getSetting("deli") + 
					"clients" + StaticValues.getSetting("deli") + "x4x"), 
					confDir);
		}
	}
	
	/**
	 * used by initialize()
	 * 
	 * @param x		sourceDir
	 * @param y		destDir
	 */
	private void copyDir(File x, File y){
		if (x.exists() && x.isDirectory()){
			if (!y.exists())
				y.mkdirs();
			if (y.isDirectory()){
				File[] entries = x.listFiles();
				for (int i = 0; i < entries.length; i++){
					if (entries[i].isDirectory())
						copyDir(entries[i], new File(y + File.separator + entries[i].getName()));
					else if (entries[i].isFile())
						copyFile(entries[i], new File(y + File.separator + entries[i].getName()));
				}
			}
		}
	}

	/**
	 * used by copyDir
	 * 
	 * @param a		sourceFile
	 * @param b		destFile
	 */
	private void copyFile(File a, File b) {
		FileChannel in, out;
		try {
			in = new FileInputStream(a).getChannel();
			out = new FileOutputStream(b).getChannel();
			in.transferTo( 0, in.size(), out);
		} catch (FileNotFoundException e) {
			Logger.getLogger(MainGUI.class).debug(e.getMessage(), e);
		} catch (IOException e) {
			Logger.getLogger(MainGUI.class).debug(e.getMessage(), e);
		}
	}
	
	private static String[] handleArgs(String[] args){
		int startAt = 0;
		String configFile = "";
		if (args.length > 0){
			if (args[0].equals("-v") || args[0].equals("--version")){
				System.out.println(version);
				return null;
			}
			else if (args[0].equals("-c") || args[0].equals("--config")){
				startAt += 2;
				configFile = args[1];
			}
			else if (args[0].equals("-h") || args[0].equals("--help")){
				System.out.println(
						"X4X - A XMMS2 client\n\n" +
						"usage X4X [options] [--clean] [filenames/playlistnames]\n\n" +
						"\t-c, --config\tset another configfile\n" +
						"\t\t(default: ${user.home}/.xmms2/clients/x4x/settings.properties)\n" +
						"\t-v, --version\tprint version\n" +
						"\t-h, --help\tprint this screen\n\n" +
						"For more infos visit http://x4x.sourceforge.net or #x4x on irc.freenode.net\n");
				return null;
			}
		}
		StaticValues.init(configFile);
		String newArgs[] = new String[args.length-startAt];
		int j = 0;
		for ( int i = startAt; i < newArgs.length; i++)
			newArgs[j++] = args[i];
		return newArgs;
	}
	
	public static void main(String[] args) {
		String x4xArgs[];
		if ((x4xArgs = handleArgs(args)) != null){
			new MainGUI(x4xArgs).open();
		}
		System.exit(0);
	}
}
