/******************************************************************
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Copyright:
 * Georg Schild (dangertools@gmail.com)
 * Franz Endstrasser (franz.endstrasser@gmail.com)
 ******************************************************************/

package gui;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Observable;
import java.util.Observer;
import java.util.Timer;
import java.util.TimerTask;

import org.Xwidgets.XLabel;
import org.Xwidgets.XSlider;
import org.apache.log4j.Logger;

import statics.StaticValues;

import core.KeepStateThread;
import core.Management;
import core.Playlist;
import core.PlaylistInformation;
import core.Title;
import core.UpdateUIType;

/**
 * Package UpdateListener which receives update from the backend and updates the UI
 */

class UpdateListener implements Observer {
	private Management myManagement;
	private boolean managementUp = false;
	private boolean guiUp = false;
	private Logger logger;
	private char roller[] = new char[]{45, 92, 124, 47};
	private int rollerPosition = 0;
	private Timer roll;
	private TimerTask rollerTask;
	private Thread plUpdater;
	
	/**
	 * Create and start the listener
	 * @param args 
	 *
	 */
	protected UpdateListener(String args[]){
		logger = Logger.getLogger(getClass());
		myManagement = Management.getInstance(this);
		handleArgs(args);
		guiIsReady();
	}
	
	/**
	 * called on startup to retreive default data from backend
	 *
	 */
	private void updateUI(){
		if (managementUp && guiUp){
			new Timer().schedule(new TimerTask(){
				public void run() {
					myManagement.updateAll();
					updatePlaylist();
					updateForPlaybackState();
					updateSliders(-1);
					myManagement.mlibGetPlaylists();
				}
			},1000);
			new Timer().schedule(new TimerTask(){
				public void run() {
					String[] query = new String[3];
					query[0] = "WHERE key='" + 
						MlibPanel.getTextFromComboIndex(MainGUI.mainGui.mlibPanel.browseBox) + "'";
					query[1] = "";
					query[2] = "own";
					myManagement.mlibSelect(query, false);
				}
			}, 8000);
		}
	}
	/**
	 * Called for handling arguments on startup
	 * 
	 * @param args
	 */
	private void handleArgs(String args[]){
		if (args != null && args.length > 0){
			int startAt = 0;
			if (args[0].equals("--clean")){
				myManagement.playlistClear();
				startAt++;
			}
			String supported[] = myManagement.getFileTypes();
			String playlists[] = myManagement.getPlaylistTypes();
			for (int i = startAt; i < args.length; i++){
				for ( int j = 0; j < supported.length; j++)
					if (args[i].toLowerCase().endsWith(supported[j]))
						myManagement.addTitleToPlaylist("file://" + args[i]);
				for ( int j = 0; j < playlists.length; j++)
					if (args[i].toLowerCase().endsWith(playlists[j])){
						myManagement.mlibLoadPlaylist("file://" + args[i], 
								args[i].substring(args[i].lastIndexOf("/")+1,args[i].indexOf(".")));
						myManagement.mlibAddPlaylistToCurrent(
								args[i].substring(args[i].lastIndexOf("/")+1,args[i].indexOf(".")));
					}
			}
		}
	}
	
	/**
	 * Set the gui to a ready state and get startupData from Management
	 *
	 */
	protected void guiIsReady(){
		guiUp = true;
		while(!myManagement.upAndRunning()){
			try {
				Thread.sleep(100);
			} catch (InterruptedException e) {
				logger.debug(e.getMessage(), e);
			}
		}
		roll = new Timer();
		managementUp = true;
		updateUI();
	}

	/**
	 * Receive updates from the Backend
	 */
	public void update(Observable arg0, Object arg1) {
		UpdateUIType uit = (UpdateUIType)arg1;
		if (managementUp && guiUp){
			switch (uit.getType()){
				case UpdateUIType.PLAYLIST_CHANGED:
					updatePlaylist();
					break;
				case UpdateUIType.PLAYBACK_ID_CHANGED:
					
					break;
				case UpdateUIType.PLAYLIST_CURRENT_POSITION_CHANGED:
					MainGUI.mainGui.playlistPanel.table.setSelectedIndex(myManagement.getPlaylistIndex());
					updateForPlaybackState();
					break;
				case UpdateUIType.PLAYBACK_STATUS_CHANGED:
					updateForPlaybackState();
					break;
				case UpdateUIType.CONFIGVAL_CHANGED:
					MainGUI.mainGui.prefGui.update();
					break;
				case UpdateUIType.PLAYTIME_CHANGED:
					int min, sec;
					min = ((Double)uit.getVal()).intValue()/60;
					sec = ((Double)uit.getVal()).intValue()%60;
					if (MainGUI.mainGui.playtime != null){
						MainGUI.mainGui.playtime.setPosition(min*60+sec);
						MainGUI.mainGui.playtime.setTooltipText(min + ":" + ((sec<10)?"0":"") + sec);
					}
					MainGUI.mainGui.buttonPanel.lcd.setNumber(min+0.01*sec);
					break;
				case UpdateUIType.VOLUME_OR_EQUALIZER_CHANGED:
					updateSliders(((Integer)uit.getVal()).intValue());
					break;
				case UpdateUIType.BACKEND_LOST:
					MainGUI.mainGui.playtime.setMax(0);
					MainGUI.mainGui.playtime.setPosition(0);
					MainGUI.mainGui.playtime.setTooltipText("0:00");
					MainGUI.mainGui.buttonPanel.lcd.setNumber(0);
					MainGUI.mainGui.tray.setTooltipText("X4X");
					updateFlow(StaticValues.getLangString("backendDisconnected"));
					updatePlaylist();
					break;
				case UpdateUIType.MLIB_LIST:
					List entries = (List)uit.getVal();
					ArrayList l = new ArrayList();
					for (ListIterator it = entries.listIterator(); it.hasNext();){
						Title t = (Title)it.next();
						if (!t.getAttribute(MlibPanel.getTextFromComboIndex(MainGUI.mainGui.mlibPanel.browseBox)).equals("") && 
								!l.contains(t.getAttribute(MlibPanel.getTextFromComboIndex(MainGUI.mainGui.mlibPanel.browseBox))))
							l.add(t.getAttribute(MlibPanel.getTextFromComboIndex(MainGUI.mainGui.mlibPanel.browseBox)));						
					}
					Collections.sort(l);
					MainGUI.mainGui.mlibPanel.browseTable.addList(l);
					break;
				case UpdateUIType.MLIB_SEARCH:
					entries = (List)uit.getVal();
					l = new ArrayList();
					ArrayList list = new ArrayList();
					for (ListIterator it = entries.listIterator(); it.hasNext();){
						Title t = (Title)it.next();
						if (!l.contains(t))
							l.add(t);
					}
					Collections.sort(l);
					for (int i = 0; i < l.size(); i++){
						Title t = (Title)l.get(i);
						list.add(new String[]{t.getAttribute("id"), t.getAttribute("artist"),
							t.getAttribute("title"), t.getAttribute("album"), t.getAttribute("genre")});
					}
					MainGUI.mainGui.mlibPanel.searchTable.addList(list);
					if (MainGUI.mainGui.tabbedPane != null)
						MainGUI.mainGui.tabbedPane.setSelection(MainGUI.
								mainGui.tabbedPane.getIndex(MainGUI.mainGui.mlibPanel.searchPanel));
					break;
				case UpdateUIType.PLAYLIST_LOADED:
					myManagement.mlibGetPlaylists();
					break;
				case UpdateUIType.PLAYLIST_INFO:
					List pentries = (List)uit.getVal();					
					MainGUI.mainGui.playlistPanel.playlistTable.clear();
					
					for (ListIterator it = pentries.listIterator(); it.hasNext();){
						PlaylistInformation pi = (PlaylistInformation)it.next();
						MainGUI.mainGui.playlistPanel.playlistTable.addItem(new String[]{"" + pi.getId(), pi.getName(), "" + pi.getItems()});
					}
					break;
				case UpdateUIType.MLIB_ENTRY_CHANGED:
					List indeces = myManagement.getCurrentPlaylist().indecesOfID(((Integer)uit.getVal()).intValue());
					for (Iterator it = indeces.iterator(); it.hasNext();){
						Integer next = (Integer)it.next();
						if (myManagement.isTitleInCache(next.intValue())){
							Title t = myManagement.getTitleForID(next.intValue());
							MainGUI.mainGui.playlistPanel.table.setText(new String[]{
									t.getAttribute("id"), t.getAttribute("title"),
									t.getAttribute("artist"), t.getAttribute("album")}, next.intValue());
						}	
					}
					break;
				case UpdateUIType.BG_WORKING:
					updateForPlaybackState();
					break;
			}
		}
		
	}
	
	/**
	 * Update flowtext/status after getting current playbackstate
	 *
	 */
	private void updateForPlaybackState() {
		if (myManagement.getBackgroundWorkerStatus() != UpdateUIType.BG_WORKING){
			if (rollerTask != null){
				rollerTask.cancel();
				rollerTask = null;
			}
			rollerPosition = 0;
			switch (myManagement.getPlaybackStatus()){
				case KeepStateThread.PAUSED:
					updateFlow(StaticValues.getLangString("paused"));
					break;
				case KeepStateThread.PLAYING:
					updateFlow("");
					break;
				default:
					updateFlow(StaticValues.getLangString("stopped"));
				MainGUI.mainGui.playtime.setPosition(0);
				MainGUI.mainGui.playtime.setTooltipText("0:00");
				MainGUI.mainGui.buttonPanel.lcd.setNumber(0);
				break;
			}
		}
		else {
			if (rollerTask == null){
				rollerTask = new TimerTask(){
					public void run() {
						if (rollerPosition > 3)
							rollerPosition=0;
						MainGUI.mainGui.status.setText(StaticValues.getLangString("backgroundWorking").toUpperCase()
								+ "  " + roller[rollerPosition++]);
					}
				};
				roll.schedule(rollerTask, 0, 50);
			}
			// TODO have a look at that later
			/*MainGUI.mainGui.status.setText(StaticValues.getLangString("backgroundWorking").toUpperCase() + ": "
			 + (myManagement.getBackgroundWorkerStatus()-UpdateUIType.BG_WORKING)
			 + " " + StaticValues.getLangString("toGo"));*/
		}
	}
	
	/**
	 * update playlist with current from Management 
	 *
	 */
	private void updatePlaylist() {
		final Playlist pl = myManagement.getCurrentPlaylist();
		if (plUpdater != null && plUpdater.isAlive())
			plUpdater.interrupt();
		if (pl.getSize() <= 0)
			MainGUI.mainGui.playlistPanel.table.clear();
		else {
			MainGUI.mainGui.playlistPanel.table.addList(pl.getIDs());
			final ArrayList stillToGo = new ArrayList();
			for (int i = 0; i < pl.getSize(); i++){
				Title t = pl.getTitleForIndex(i);
				if (myManagement.isTitleInCache(Integer.parseInt(t.getAttribute("id")))){
					t = myManagement.getTitleForID(Integer.parseInt(t.getAttribute("id")));
					pl.updateTitleAtIndex(i, t);
					MainGUI.mainGui.playlistPanel.table.setText(new String[]{
							t.getAttribute("id"), t.getAttribute("title"),
							t.getAttribute("artist"), t.getAttribute("album")}, i);
				}
				else {
					stillToGo.add(""+i);
				}
			}
			if (stillToGo.size() > 0){
				plUpdater = new Thread(){
					public void run(){
						int ttl = 30;
						try {
							sleep(1000);
							while (ttl-- > 0 && stillToGo.size() > 0){
								sleep(500);
								for (Iterator it = stillToGo.iterator(); it.hasNext();){
									int index = Integer.parseInt(""+it.next());
									int id = Integer.parseInt(pl.getTitleForIndex(index).getAttribute("id"));
									if (myManagement.isTitleInCache(id)){
										Title t = myManagement.getTitleForID(id);
										if ( t != null){
											MainGUI.mainGui.playlistPanel.table.setText(new String[]{
													t.getAttribute("id"), t.getAttribute("title"),
													t.getAttribute("artist"), t.getAttribute("album")}, index);
											it.remove();
										}
										else sleep(40);
									}
								}
							}
						} catch (Exception e) {
							logger.debug(e.getMessage(), e);
						}
					}
				};
				plUpdater.start();
			}
		}
	}
	
	/**
	 * update sliders volume or equalizer
	 * 
	 * @param x		if -1 we'll update volume, else equalizer x
	 */
	private void updateSliders(int x){
		if ( x == -1 && MainGUI.mainGui.volume != null){
			MainGUI.mainGui.volume.setPosition(myManagement.getVolume());
			MainGUI.mainGui.volume.setTooltipText(StaticValues.getLangString("volume") + ": "+myManagement.getVolume());
		}
		else {
			if (MainGUI.mainGui.eqPanel != null){
				try {
					((XSlider)MainGUI.mainGui.eqPanel.sliders.get(x)).setPosition((myManagement.getEQ(x)-50)*(-1));
					((XLabel)MainGUI.mainGui.eqPanel.labels.get(x)).setText(""+(myManagement.getEQ(x)));
				} catch (Exception e){
					logger.debug(e.getMessage(), e);
				}
			}
		}
	}
	
	/**
	 * update flowtext with given text
	 * 
	 * @param text		if empty we'll get artist and title for current playing title, else we set text
	 */
	private void updateFlow(String text){
		if (text.equals("")){
			Title t = myManagement.getPlayingTitle();
			if (MainGUI.mainGui.flow != null)
				MainGUI.mainGui.flow.setFlowText(t.getAttribute("artist") + " - " + t.getAttribute("title"));
			if (MainGUI.mainGui.playtime != null)
				MainGUI.mainGui.playtime.setMax(t.getLength());
			int min = t.getLength()/60;
			int sec = t.getLength()%60;
			MainGUI.mainGui.tray.setTooltipText(t.getAttribute("artist") + "\n" + t.getAttribute("title") + "\n" +
					min + ":" + ((sec<10)?"0":"") + sec);
			MainGUI.mainGui.status.setText(StaticValues.getLangString("playing") + ": " + t.getAttribute("artist") + " - " + t.getAttribute("title"));
		}
		else{
			if (MainGUI.mainGui.flow != null)
				MainGUI.mainGui.flow.setFlowText(text);
			MainGUI.mainGui.status.setText(text);
		}
	}
}
