/******************************************************************
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Copyright:
 * Georg Schild (dangertools@gmail.com)
 * Franz Endstrasser (franz.endstrasser@gmail.com)
 ******************************************************************/

package tests;

import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Observable;
import java.util.Observer;

import core.Management;
import core.PlaylistInformation;
import core.ReturnValue;
import core.Title;
import core.UpdateUIType;
import junit.framework.TestCase;

public class MedialibTests extends TestCase implements Observer{
	private Management m = AllTests.m;
	private LinkedList playlistInformations;
	private ReturnValue ret;
	
	protected void setUp() throws Exception {
		m = Management.getInstance(this);
		playlistInformations = new LinkedList();
	}
	
	public void testNewDelPlaylist() {
		String name = "testPlaylist";
		PlaylistInformation result;
	
		/* remove playlist, if the mlib contains it already */
		removePlaylist(name);

		/* add it */
		m.waitForID(m.mlibNewPlaylist(name));
		
		result = searchPlaylist(name);
		
		/* test if it contains the new playlist */
		assertEquals(result.getName(), name);
		
		/* remove the new playlist */
		removePlaylist(name);
		
		result = searchPlaylist(name);
		
		/* test if it doesn't contain the new playlist any longer*/
		assertEquals(result, null);
	}

	public void testSavePlaylist() {
		String name = "testPlaylist";
		PlaylistInformation result;
		
		m.waitForID(m.playlistClear());
		
		/* remove playlist, if the mlib contains it already */
		removePlaylist(name);
		
		/* add some titles for testing */
		addTestTitles();
		
		m.waitForID(m.mlibSaveCurrentPlaylist(name));
		result = searchPlaylist(name);
		
		assertEquals(result.getName(),name);
		m.waitForID(m.playlistClear());
	}
	
	
	public void testRenamePlaylist(){
		String name = "testPlaylist";
		String newname = "newTestPlaylist";
		PlaylistInformation result;
		
		/* remove playlist, if the mlib contains it already */
		removePlaylist(name);
		removePlaylist(newname);
		
		/* add it */
		m.waitForID(m.mlibNewPlaylist(name));

		result = searchPlaylist(name);
		
		/* test if it contains the new playlist */
		assertEquals(result.getName(), name);
		
		/* rename it */
		m.waitForID(m.mlibRenamePlaylist(name, newname));
		
		/* search for the new name */
		result = searchPlaylist(newname);
		
		/* test if we found the new name*/
		assertEquals(result.getName(),newname);
		
		/* remove the new playlist */
		removePlaylist(newname);
		
		result = searchPlaylist(newname);
		
		/* test if it doesn't contain the new playlist any longer*/
		assertEquals(result, null);
	}
	


	private LinkedList addTestTitles() {
		LinkedList addedTitles = new LinkedList();
		LinkedList entries;

		ret = m.waitForID(m.mlibSelect(new String[]{"WHERE key='artist' AND value LIKE '%'","", "own"}, true));
		assertEquals(ReturnValue.MEDIALIB_SEARCH_TYPE,ret.getType());
		
		UpdateUIType uit = new UpdateUIType(UpdateUIType.MLIB_SEARCH, ret.getValue());
		entries = (LinkedList)uit.getVal();
		for (int i=0; (i < 10) && ( i < entries.size()); i++){
			Title t = (Title)entries.get(i);
			if (!addedTitles.contains(t)) {
				ret = m.waitForID(m.addIdToPlaylist(Integer.parseInt(t.getAttribute("id"))));
				assertEquals(ret.getType(), ReturnValue.VOID_TYPE);
				addedTitles.add(t);
			}
		}
		
		return addedTitles;
	}
	
	private void removePlaylist(String name) {
		PlaylistInformation result;
		result = searchPlaylist(name);
		if (result != null) {
			/* remove playlist */
			ret = m.waitForID(m.mlibRemovePlaylist(name));
			assertEquals(ret.getType(), ReturnValue.VOID_TYPE);
		}		
	}
	
	private void updateLocalPlaylists(UpdateUIType uit) {
		/* update list of Playlistinfos */
		List pentries = (List)uit.getVal();					
		playlistInformations = new LinkedList();
		
		for (ListIterator it = pentries.listIterator(); it.hasNext();){
			PlaylistInformation pi = (PlaylistInformation)it.next();
			playlistInformations.add(pi);
		}			
	}
	
	private PlaylistInformation searchPlaylist(String name) {
		PlaylistInformation piret = null;
		
		/* get playlists */
		ret = m.waitForID(m.mlibGetPlaylists());
		assertEquals(ret.getType(), ReturnValue.PLAYLISTINFORMATION_TYPE);
		
		UpdateUIType uit = new UpdateUIType(UpdateUIType.PLAYLIST_INFO, ret.getValue());
		updateLocalPlaylists(uit);
			
		for (ListIterator it = playlistInformations.listIterator(); it.hasNext();){
			PlaylistInformation pi = (PlaylistInformation)it.next();
			if (pi.getName().equals(name))
				piret = pi;
		}	
		return piret;	
	}
	
	public void update(Observable arg0, Object arg1) {
		UpdateUIType uit = (UpdateUIType)arg1;
		switch (uit.getType()){
			case UpdateUIType.PLAYLIST_CHANGED:
				break;
			case UpdateUIType.PLAYLIST_INFO:
				List pentries = (List)uit.getVal();					
				playlistInformations = new LinkedList();
				
				for (ListIterator it = pentries.listIterator(); it.hasNext();){
					PlaylistInformation pi = (PlaylistInformation)it.next();
					playlistInformations.add(pi);
				}
		}
	}
	
}
